import time
import os
import shutil
import logging
import sqlite3
import datetime

import pandas

import gtfspy
from gtfspy import util
from gtfspy.gtfs import GTFS
from gtfspy.import_loaders.day_loader import recreate_days_table
from gtfspy.import_loaders.day_trips_materializer import recreate_day_trips2_table
from gtfspy.import_loaders.stop_times_loader import resequence_stop_times_seq_values
from gtfspy.import_loaders.trip_loader import update_trip_travel_times_ds
from gtfspy.util import wgs84_distance, set_process_timezone
from gtfspy import stats
from gtfspy import gtfs

FILTERED = True
NOT_FILTERED = False

DELETE_FREQUENCIES_NOT_REFERENCED_IN_TRIPS_SQL = "DELETE FROM frequencies WHERE trip_I NOT IN (SELECT DISTINCT trip_I FROM trips)"
DELETE_SHAPES_NOT_REFERENCED_IN_TRIPS_SQL = 'DELETE FROM shapes WHERE shape_id NOT IN (SELECT shape_id FROM trips)'
DELETE_ROUTES_NOT_PRESENT_IN_TRIPS_SQL = 'DELETE FROM routes WHERE route_I NOT IN (SELECT route_I FROM trips)'
DELETE_DAYS_ENTRIES_NOT_PRESENT_IN_TRIPS_SQL = "DELETE FROM days WHERE trip_I NOT IN (SELECT trip_I FROM trips)"
DELETE_DAY_TRIPS2_ENTRIES_NOT_PRESENT_IN_TRIPS_SQL = "DELETE FROM day_trips2 WHERE trip_I NOT IN (SELECT trip_I FROM trips)"
DELETE_FREQUENCIES_ENTRIES_NOT_PRESENT_IN_TRIPS = "DELETE FROM frequencies WHERE trip_I NOT IN (SELECT trip_I FROM trips)"
DELETE_CALENDAR_ENTRIES_FOR_NON_REFERENCE_SERVICE_IS_SQL = "DELETE FROM calendar WHERE service_I NOT IN (SELECT distinct(service_I) FROM trips)"
DELETE_CALENDAR_DATES_ENTRIES_FOR_NON_REFERENCE_SERVICE_IS_SQL = "DELETE FROM calendar_dates WHERE service_I NOT IN (SELECT distinct(service_I) FROM trips)"
DELETE_AGENCIES_NOT_REFERENCED_IN_ROUTES_SQL = "DELETE FROM agencies WHERE agency_I NOT IN (SELECT distinct(agency_I) FROM routes)"
DELETE_STOP_TIMES_NOT_REFERENCED_IN_TRIPS_SQL = 'DELETE FROM stop_times WHERE trip_I NOT IN (SELECT trip_I FROM trips)'
DELETE_STOP_DISTANCE_ENTRIES_WITH_NONEXISTENT_STOPS_SQL = "DELETE FROM stop_distances " \
                                                          "WHERE from_stop_I NOT IN (SELECT stop_I FROM stops) " \
                                                          " OR to_stop_I NOT IN (SELECT stop_I FROM stops)"
DELETE_TRIPS_NOT_IN_DAYS_SQL = 'DELETE FROM trips WHERE trip_I NOT IN (SELECT trip_I FROM days)'
DELETE_TRIPS_NOT_REFERENCED_IN_STOP_TIMES = 'DELETE FROM trips WHERE trip_I NOT IN (SELECT trip_I FROM stop_times)'


class FilterExtract(object):

    def __init__(self,
                 G,
                 copy_db_path,
                 buffer_distance_km=None,
                 buffer_lat=None,
                 buffer_lon=None,
                 update_metadata=True,
                 start_date=None,
                 end_date=None,
                 agency_ids_to_preserve=None,
                 agency_distance=None):
        """
        Copy a database, and then based on various filters.
        Only method `create_filtered_copy` is provided as we do not want to take the risk of
        losing the data stored in the original database.

        G: gtfspy.gtfs.GTFS
            the original database
        copy_db_path : str
            path to another database database
        update_metadata : boolean, optional
            whether to update metadata of the feed, defaulting to true
            (this option is mainly available for testing purposes)
        start_date : str, or datetime.datetime
            filter out all data taking place before end_date (the start_time_ut of the end date)
            Date format "YYYY-MM-DD"
            (end_date_ut is not included after filtering)
        end_date : str, or datetime.datetime
            Filter out all data taking place after end_date
            The end_date is not included after filtering.
        agency_ids_to_preserve : iterable
            List of agency_ids to retain (str) (e.g. 'HSL' for Helsinki)
            Only routes by the listed agencies are then considered
        agency_distance : float
            Only evaluated in combination with agency filter.
            Distance (in km) to the other near-by stops that should be included in addition to
            the ones defined by the agencies.
            All vehicle trips going through at least two such stops would then be included in the
            export. Note that this should not be a recursive thing.
            Or should it be? :)
        buffer_lat : float
            Latitude of the buffer zone center
        buffer_lon : float
            Longitude of the buffer zone center
        buffer_distance : float
            Distance from the buffer zone center (in kilometers)

        Returns
        -------
        None
        """
        if start_date and end_date:
            if isinstance(start_date, (datetime.datetime, datetime.date)):
                self.start_date = start_date.strftime("%Y-%m-%d")
            else:
                self.start_date = start_date
            if isinstance(end_date, (datetime.datetime, datetime.date)):
                end_date_dt = end_date
                self.end_date = end_date.strftime("%Y-%m-%d")
            else:
                self.end_date = end_date
                end_date_dt = datetime.datetime.strptime(self.end_date, "%Y-%m-%d")
            end_date_to_include = end_date_dt - datetime.timedelta(days=1)
            self.end_date_to_include_str = end_date_to_include.strftime("%Y-%m-%d")
        else:
            self.start_date = None
            self.end_date = None
        self.copy_db_conn = None
        self.copy_db_path = copy_db_path

        self.agency_ids_to_preserve = agency_ids_to_preserve
        self.gtfs = G
        self.buffer_lat = buffer_lat
        self.buffer_lon = buffer_lon
        self.buffer_distance_km = buffer_distance_km
        self.update_metadata = update_metadata

        if agency_distance is not None:
            raise NotImplementedError

        self.this_db_path = self.gtfs.get_main_database_path()
        assert os.path.exists(self.this_db_path), "Copying of in-memory databases is not supported"
        assert os.path.exists(os.path.dirname(os.path.abspath(copy_db_path))), \
            "the directory where the copied database will reside should exist beforehand"
        assert not os.path.exists(copy_db_path), "the resulting database exists already: %s" % copy_db_path

    def create_filtered_copy(self):
        # this with statement
        # is used to ensure that no corrupted/uncompleted files get created in case of problems
        with util.create_file(self.copy_db_path) as tempfile:
            logging.info("copying database")
            shutil.copy(self.this_db_path, tempfile)
            self.copy_db_conn = sqlite3.connect(tempfile)
            assert isinstance(self.copy_db_conn, sqlite3.Connection)

            filtered = False
            filtered = self._delete_rows_by_start_and_end_date() or filtered
            if self.copy_db_conn.execute('SELECT count(*) FROM days').fetchone() == (0,):
                raise ValueError('No data left after filtering')
            filtered = self._filter_by_calendar() or filtered
            filtered = self._filter_by_agency() or filtered
            filtered = self._filter_spatially() or filtered
            self.copy_db_conn.commit()
            if filtered:
                update_secondary_data_copies(db_conn=self.copy_db_conn)
            if self.update_metadata:
                self._update_metadata()
        return

    def _delete_rows_by_start_and_end_date(self):
        """
        Removes rows from the sqlite database copy that are out of the time span defined by start_date and end_date
        :param gtfs: GTFS object
        :param copy_db_conn: sqlite database connection
        :param start_date:
        :param end_date:
        :return:
        """
        # filter by start_time_ut and end_date_ut:
        if (self.start_date is not None) and (self.end_date is not None):
            start_date_ut = self.gtfs.get_day_start_ut(self.start_date)
            end_date_ut = self.gtfs.get_day_start_ut(self.end_date)
            if self.copy_db_conn.execute("SELECT count(*) FROM day_trips2 WHERE start_time_ut IS null "
                                         "OR end_time_ut IS null").fetchone() != (0,):
                raise ValueError("Missing information in day_trips2 (start_time_ut and/or end_time_ut), "
                                 "check trips.start_time_ds and trips.end_time_ds.")
            logging.info("Filtering based on start_time_ut and end_time_ut")
            table_to_preserve_map = {
                "calendar": "start_date < date({filter_end_ut}, 'unixepoch', 'localtime') "
                            "AND "
                            "end_date >= date({filter_start_ut}, 'unixepoch', 'localtime') ",
                "calendar_dates": "date >= date({filter_start_ut}, 'unixepoch', 'localtime') "
                                  "AND "
                                  "date < date({filter_end_ut}, 'unixepoch', 'localtime') ",
                "day_trips2": 'start_time_ut < {filter_end_ut} '
                              'AND '
                              'end_time_ut > {filter_start_ut} ',
                "days": "day_start_ut >= {filter_start_ut} "
                        "AND "
                        "day_start_ut < {filter_end_ut} "
                }
            table_to_remove_map = {key: "WHERE NOT ( " + to_preserve + " );"
                                   for key, to_preserve in table_to_preserve_map.items() }

            # Ensure that process timezone is correct as we rely on 'localtime' in the SQL statements.
            GTFS(self.copy_db_conn).set_current_process_time_zone()
            # remove the 'source' entries from tables
            for table, query_template in table_to_remove_map.items():
                param_dict = {"filter_start_ut": str(start_date_ut),
                              "filter_end_ut": str(end_date_ut)}
                query = "DELETE FROM " + table + " " + \
                        query_template.format(**param_dict)
                self.copy_db_conn.execute(query)
                self.copy_db_conn.commit()

            return FILTERED
        else:
            return NOT_FILTERED

    def _soft_filter_by_calendar(self):
        pass
        """
        # TODO: soft filtering, where the recursive deletion of rows depends on the initially removed rows, not on missing links between ID fields.
        The reason for doing this is to detect unreferenced rows that possibly should be included in the filtered extract.
        This gives a much more accurate perspective on the quality of the feed


        PSeudocode:
        select all trip_I not to be deleted based on dates
        select all routes not to be deleted based on trip_I
        delete all agencies where agency_I not in routes to be saved nor is unreferenced in routes
        delete all routes where route_I not in trips to be saved
        delete all

        delete = {'agency': 'DELETE FROM agencies WHERE NOT agency_I IN',
                  'routes': 'DELETE FROM routes WHERE NOT route_I IN',
                  'trips': 'DELETE FROM trips WHERE NOT trip_I IN',
                  'stops': 'DELETE FROM stops WHERE NOT stop_I IN',
                  'stop_times': 'DELETE FROM stop_times WHERE NOT trip_I IN',
                  'days': 'DELETE trip_I FROM days WHERE NOT ({start_ut} <= day_start_ut AND day_start_ut < {end_ut})'}
        select = {'routes': 'SELECT agency_I FROM routes WHERE route_I IN',
                  'trips': 'SELECT route_I FROM trips WHERE trip_I IN',
                  'stop_times': 'SELECT stop_I FROM stop_times WHERE trip_I IN',
                  'days': 'SELECT trip_I FROM days WHERE ({start_ut} <= day_start_ut AND day_start_ut < {end_ut})'}
        'SELECT agency_I FROM agencies LEFT JOIN routes WHERE routes.'
        query = delete['agency']+select['routes']+select['trips']+select['days']

        agency_query = 'DELETE FROM agencies WHERE NOT agency_I IN (SELECT agency_I FROM routes WHERE route_I IN (SELECT route_I FROM trips WHERE trip_I IN ())) AND '
        """
        pass

    def _filter_by_calendar(self):
        """
        update calendar table's services
        :param copy_db_conn:
        :param start_date:
        :param end_date:
        :return:
        """
        if (self.start_date is not None) and (self.end_date is not None):
            logging.info("Making date extract")

            start_date_query = "UPDATE calendar " \
                               "SET start_date='{start_date}' " \
                               "WHERE start_date<'{start_date}' ".format(start_date=self.start_date)
            self.copy_db_conn.execute(start_date_query)

            end_date_query = "UPDATE calendar " \
                             "SET end_date='{end_date_to_include}' " \
                             "WHERE end_date>'{end_date_to_include}' " \
                .format(end_date_to_include=self.end_date_to_include_str)
            self.copy_db_conn.execute(end_date_query)

            # then recursively delete further data:
            self.copy_db_conn.execute(DELETE_TRIPS_NOT_IN_DAYS_SQL)
            self.copy_db_conn.execute(DELETE_SHAPES_NOT_REFERENCED_IN_TRIPS_SQL)
            self.copy_db_conn.execute(DELETE_STOP_TIMES_NOT_REFERENCED_IN_TRIPS_SQL)
            delete_stops_not_in_stop_times_and_not_as_parent_stop(self.copy_db_conn)
            self.copy_db_conn.execute(DELETE_STOP_DISTANCE_ENTRIES_WITH_NONEXISTENT_STOPS_SQL)
            self.copy_db_conn.execute(DELETE_ROUTES_NOT_PRESENT_IN_TRIPS_SQL)
            self.copy_db_conn.execute(DELETE_AGENCIES_NOT_REFERENCED_IN_ROUTES_SQL)
            self.copy_db_conn.commit()
            return FILTERED
        else:
            return NOT_FILTERED

    def _filter_by_agency(self):
        """
        filter by agency ids
        :param copy_db_conn:
        :param agency_ids_to_preserve:
        :return:
        """
        if self.agency_ids_to_preserve is not None:
            logging.info("Filtering based on agency_ids")
            agency_ids_to_preserve = list(self.agency_ids_to_preserve)
            agencies = pandas.read_sql("SELECT * FROM agencies", self.copy_db_conn)
            agencies_to_remove = []
            for idx, row in agencies.iterrows():
                if row['agency_id'] not in agency_ids_to_preserve:
                    agencies_to_remove.append(row['agency_id'])
            for agency_id in agencies_to_remove:
                self.copy_db_conn.execute('DELETE FROM agencies WHERE agency_id=?', (agency_id,))
            # and remove recursively related to the agencies:
            self.copy_db_conn.execute('DELETE FROM routes WHERE '
                                      'agency_I NOT IN (SELECT agency_I FROM agencies)')
            self.copy_db_conn.execute('DELETE FROM trips WHERE '
                                      'route_I NOT IN (SELECT route_I FROM routes)')
            self.copy_db_conn.execute('DELETE FROM calendar WHERE '
                                      'service_I NOT IN (SELECT service_I FROM trips)')
            self.copy_db_conn.execute('DELETE FROM calendar_dates WHERE '
                                      'service_I NOT IN (SELECT service_I FROM trips)')
            self.copy_db_conn.execute('DELETE FROM days WHERE '
                                      'trip_I NOT IN (SELECT trip_I FROM trips)')
            self.copy_db_conn.execute('DELETE FROM stop_times WHERE '
                                      'trip_I NOT IN (SELECT trip_I FROM trips)')
            self.copy_db_conn.execute('DELETE FROM stop_times WHERE '
                                      'trip_I NOT IN (SELECT trip_I FROM trips)')
            self.copy_db_conn.execute('DELETE FROM shapes WHERE '
                                      'shape_id NOT IN (SELECT shape_id FROM trips)')
            self.copy_db_conn.execute('DELETE FROM day_trips2 WHERE '
                                      'trip_I NOT IN (SELECT trip_I FROM trips)')
            self.copy_db_conn.commit()
            return FILTERED
        else:
            return NOT_FILTERED

    def _filter_spatially(self):
        """
        Filter the feed based on self.buffer_distance_km from self.buffer_lon and self.buffer_lat.

        1. First include all stops that are within self.buffer_distance_km from self.buffer_lon and self.buffer_lat.
        2. Then include all intermediate stops that are between any of the included stop pairs with some PT trip.
        3. Repeat step 2 until no more stops are to be included.

        As a summary this process should get rid of PT network tendrils, but should preserve the PT network intact
        at its core.
        """
        if self.buffer_lat is None or self.buffer_lon is None or self.buffer_distance_km is None:
            return NOT_FILTERED

        print("filtering with lat: " + str(self.buffer_lat) +
              " lon: " + str(self.buffer_lon) +
              " buffer distance: " + str(self.buffer_distance_km))
        remove_all_trips_fully_outside_buffer(self.copy_db_conn,
                                              self.buffer_lat,
                                              self.buffer_lon,
                                              self.buffer_distance_km,
                                              update_secondary_data=False)
        logging.info("Making spatial extract")

        find_distance_func_name = add_wgs84_distance_function_to_db(self.copy_db_conn)
        assert find_distance_func_name == "find_distance"

        # select all stops that are within the buffer and have some stop_times assigned.
        stop_distance_filter_sql_base = (
            "SELECT DISTINCT stops.stop_I FROM stops, stop_times" +
            "    WHERE CAST(find_distance(lat, lon, {buffer_lat}, {buffer_lon}) AS INT) < {buffer_distance_meters}" +
            "     AND stops.stop_I=stop_times.stop_I"
        )
        stops_within_buffer_sql = stop_distance_filter_sql_base.format(
            buffer_lat=float(self.buffer_lat),
            buffer_lon=float(self.buffer_lon),
            buffer_distance_meters=int(self.buffer_distance_km * 1000)
        )
        stops_within_buffer = set(row[0] for row in self.copy_db_conn.execute(stops_within_buffer_sql))

        # For each trip_I, find smallest (min_seq) and largest (max_seq) stop sequence numbers that
        # are within the soft buffer_distance from the buffer_lon and buffer_lat, and add them into the
        # list of stops to preserve.
        # Note that if a trip is OUT-IN-OUT-IN-OUT, this process preserves (at least) the part IN-OUT-IN of the trip.
        # Repeat until no more stops are found.

        stops_within_buffer_string = "(" +",".join(str(stop_I) for stop_I in stops_within_buffer) +  ")"
        trip_min_max_include_seq_sql =  (
            'SELECT trip_I, min(seq) AS min_seq, max(seq) AS max_seq FROM stop_times, stops '
                    'WHERE stop_times.stop_I = stops.stop_I '
                    ' AND stops.stop_I IN {stop_I_list}'
                    ' GROUP BY trip_I'
        ).format(stop_I_list=stops_within_buffer_string)
        trip_I_min_seq_max_seq_df = pandas.read_sql(trip_min_max_include_seq_sql, self.copy_db_conn)

        for trip_I_seq_row in trip_I_min_seq_max_seq_df.itertuples():
            trip_I = trip_I_seq_row.trip_I
            min_seq = trip_I_seq_row.min_seq
            max_seq = trip_I_seq_row.max_seq
            # DELETE FROM STOP_TIMES
            if min_seq == max_seq:
                # Only one entry in stop_times to be left, remove whole trip.
                self.copy_db_conn.execute("DELETE FROM stop_times WHERE trip_I={trip_I}".format(trip_I=trip_I))
                self.copy_db_conn.execute("DELETE FROM trips WHERE trip_i={trip_I}".format(trip_I=trip_I))
            else:
                # DELETE STOP_TIME ENTRIES BEFORE ENTERING AND AFTER DEPARTING THE BUFFER AREA
                DELETE_STOP_TIME_ENTRIES_SQL = \
                    "DELETE FROM stop_times WHERE trip_I={trip_I} AND (seq<{min_seq} OR seq>{max_seq})"\
                    .format(trip_I=trip_I, max_seq=max_seq, min_seq=min_seq)
                self.copy_db_conn.execute(DELETE_STOP_TIME_ENTRIES_SQL)

                STOPS_NOT_WITHIN_BUFFER__FOR_TRIP_SQL = \
                    "SELECT seq, stop_I IN {stops_within_hard_buffer} AS within FROM stop_times WHERE trip_I={trip_I} ORDER BY seq"\
                    .format(stops_within_hard_buffer=stops_within_buffer_string, trip_I=trip_I)
                stop_times_within_buffer_df = pandas.read_sql(STOPS_NOT_WITHIN_BUFFER__FOR_TRIP_SQL, self.copy_db_conn)
                if stop_times_within_buffer_df['within'].all():
                    continue
                else:
                    _split_trip(self.copy_db_conn, trip_I, stop_times_within_buffer_df)


        # Delete all shapes that are not fully within the buffer to avoid shapes going outside
        # the buffer area in a some cases.
        # This could probably be done in some more sophisticated way though (per trip)
        SHAPE_IDS_NOT_WITHIN_BUFFER_SQL = \
            "SELECT DISTINCT shape_id FROM SHAPES " \
            "WHERE CAST(find_distance(lat, lon, {buffer_lat}, {buffer_lon}) AS INT) > {buffer_distance_meters}" \
            .format(buffer_lat=self.buffer_lat,
                    buffer_lon=self.buffer_lon,
                    buffer_distance_meters=self.buffer_distance_km * 1000)
        DELETE_ALL_SHAPE_IDS_NOT_WITHIN_BUFFER_SQL = "DELETE FROM shapes WHERE shape_id IN (" \
                                                          + SHAPE_IDS_NOT_WITHIN_BUFFER_SQL + ")"
        self.copy_db_conn.execute(DELETE_ALL_SHAPE_IDS_NOT_WITHIN_BUFFER_SQL)
        SET_SHAPE_ID_TO_NULL_FOR_HARD_BUFFER_FILTERED_SHAPE_IDS = \
            "UPDATE trips SET shape_id=NULL WHERE trips.shape_id IN (" + SHAPE_IDS_NOT_WITHIN_BUFFER_SQL + ")"
        self.copy_db_conn.execute(SET_SHAPE_ID_TO_NULL_FOR_HARD_BUFFER_FILTERED_SHAPE_IDS)


        # Delete trips with only one stop
        self.copy_db_conn.execute('DELETE FROM stop_times WHERE '
                                  'trip_I IN (SELECT trip_I FROM '
                                  '(SELECT trip_I, count(*) AS N_stops from stop_times '
                                  'GROUP BY trip_I) q1 '
                                  'WHERE N_stops = 1)')

        # Delete trips with only one stop but several instances in stop_times
        self.copy_db_conn.execute('DELETE FROM stop_times WHERE '
                                  'trip_I IN (SELECT q1.trip_I AS trip_I FROM '
                                    '(SELECT trip_I, stop_I, count(*) AS stops_per_stop FROM stop_times '
                                    'GROUP BY trip_I, stop_I) q1, '
                                    '(SELECT trip_I, count(*) as n_stops FROM stop_times '
                                    'GROUP BY trip_I) q2 '
                                    'WHERE q1.trip_I = q2.trip_I AND n_stops = stops_per_stop)')

        # Delete all stop_times for uncovered stops
        delete_stops_not_in_stop_times_and_not_as_parent_stop(self.copy_db_conn)
        # Consecutively delete all the rest remaining.
        self.copy_db_conn.execute(DELETE_TRIPS_NOT_REFERENCED_IN_STOP_TIMES)
        self.copy_db_conn.execute(DELETE_ROUTES_NOT_PRESENT_IN_TRIPS_SQL)
        self.copy_db_conn.execute(DELETE_AGENCIES_NOT_REFERENCED_IN_ROUTES_SQL)
        self.copy_db_conn.execute(DELETE_SHAPES_NOT_REFERENCED_IN_TRIPS_SQL)
        self.copy_db_conn.execute(DELETE_STOP_DISTANCE_ENTRIES_WITH_NONEXISTENT_STOPS_SQL)
        self.copy_db_conn.execute(DELETE_FREQUENCIES_ENTRIES_NOT_PRESENT_IN_TRIPS)
        remove_dangling_shapes(self.copy_db_conn)
        self.copy_db_conn.commit()
        return FILTERED

    def _update_metadata(self):
        # Update metadata
        G_orig = self.gtfs
        if self.update_metadata:
            print("Updating metadata")
            logging.info("Updating metadata")
            G_copy = gtfs.GTFS(self.copy_db_conn)
            G_copy.meta['copied_from'] = self.this_db_path
            G_copy.meta['copy_time_ut'] = time.time()
            G_copy.meta['copy_time'] = time.ctime()

            # Copy some keys directly.
            try:
                for key in ['original_gtfs',
                            'download_date',
                            'location_name',
                            'timezone', ]:
                    G_copy.meta[key] = G_orig.meta[key]
            # This part is for gtfs objects with multiple sources
            except:
                for k, v in G_copy.meta.items():
                    if 'feed_' in k:
                        G_copy.meta[k] = G_orig.meta[k]
                for key in ['location_name',
                            'timezone', ]:
                    G_copy.meta[key] = G_orig.meta[key]
            # Update *all* original metadata under orig_ namespace.
            G_copy.meta.update(('orig_' + k, v) for k, v in G_orig.meta.items())

            stats.update_stats(G_copy)

            # print "Vacuuming..."
            self.copy_db_conn.execute('VACUUM;')
            # print "Analyzing..."
            self.copy_db_conn.execute('ANALYZE;')
            self.copy_db_conn.commit()
        return

def delete_stops_not_in_stop_times_and_not_as_parent_stop(conn):
    _STOPS_REFERENCED_IN_STOP_TIMES_OR_AS_PARENT_STOP_I_SQL = \
        "SELECT DISTINCT stop_I FROM stop_times " \
        "UNION " \
        "SELECT DISTINCT parent_I as stop_I FROM stops WHERE parent_I IS NOT NULL"
    DELETE_STOPS_NOT_REFERENCED_IN_STOP_TIMES_AND_NOT_PARENT_STOP_SQL = \
        "DELETE FROM stops WHERE stop_I NOT IN (" + \
        _STOPS_REFERENCED_IN_STOP_TIMES_OR_AS_PARENT_STOP_I_SQL + ")"
    # It is possible that there is some "parent_I" recursion going on, and thus we
    # execute the same SQL query three times.
    conn.execute(DELETE_STOPS_NOT_REFERENCED_IN_STOP_TIMES_AND_NOT_PARENT_STOP_SQL)
    conn.execute(DELETE_STOPS_NOT_REFERENCED_IN_STOP_TIMES_AND_NOT_PARENT_STOP_SQL)
    conn.execute(DELETE_STOPS_NOT_REFERENCED_IN_STOP_TIMES_AND_NOT_PARENT_STOP_SQL)

def add_wgs84_distance_function_to_db(conn):
    function_name = "find_distance"
    conn.create_function(function_name, 4, wgs84_distance)
    return function_name


def remove_all_trips_fully_outside_buffer(db_conn, center_lat, center_lon, buffer_km, update_secondary_data=True):
    """
    Not used in the regular filter process for the time being.

    Parameters
    ----------
    db_conn: sqlite3.Connection
        connection to the GTFS object
    center_lat: float
    center_lon: float
    buffer_km: float
    """
    distance_function_str = add_wgs84_distance_function_to_db(db_conn)
    stops_within_buffer_query_sql = "SELECT stop_I FROM stops WHERE CAST(" + distance_function_str + \
                                "(lat, lon, {lat} , {lon}) AS INT) < {d_m}"\
        .format(lat=float(center_lat), lon=float(center_lon), d_m=int(1000*buffer_km))
    select_all_trip_Is_where_stop_I_is_within_buffer_sql = "SELECT distinct(trip_I) FROM stop_times WHERE stop_I IN (" + stops_within_buffer_query_sql + ")"
    trip_Is_to_remove_sql = "SELECT trip_I FROM trips WHERE trip_I NOT IN ( " + select_all_trip_Is_where_stop_I_is_within_buffer_sql + ")"
    trip_Is_to_remove = pandas.read_sql(trip_Is_to_remove_sql, db_conn)["trip_I"].values
    trip_Is_to_remove_string = ",".join([str(trip_I) for trip_I in trip_Is_to_remove])
    remove_all_trips_fully_outside_buffer_sql = "DELETE FROM trips WHERE trip_I IN (" + trip_Is_to_remove_string + ")"
    remove_all_stop_times_where_trip_I_fully_outside_buffer_sql = "DELETE FROM stop_times WHERE trip_I IN (" + trip_Is_to_remove_string  + ")"
    db_conn.execute(remove_all_trips_fully_outside_buffer_sql)
    db_conn.execute(remove_all_stop_times_where_trip_I_fully_outside_buffer_sql)
    delete_stops_not_in_stop_times_and_not_as_parent_stop(db_conn)
    db_conn.execute(DELETE_ROUTES_NOT_PRESENT_IN_TRIPS_SQL)
    db_conn.execute(DELETE_SHAPES_NOT_REFERENCED_IN_TRIPS_SQL)
    db_conn.execute(DELETE_DAYS_ENTRIES_NOT_PRESENT_IN_TRIPS_SQL)
    db_conn.execute(DELETE_DAY_TRIPS2_ENTRIES_NOT_PRESENT_IN_TRIPS_SQL)
    db_conn.execute(DELETE_CALENDAR_ENTRIES_FOR_NON_REFERENCE_SERVICE_IS_SQL)
    db_conn.execute(DELETE_CALENDAR_DATES_ENTRIES_FOR_NON_REFERENCE_SERVICE_IS_SQL)
    db_conn.execute(DELETE_FREQUENCIES_ENTRIES_NOT_PRESENT_IN_TRIPS)
    db_conn.execute(DELETE_AGENCIES_NOT_REFERENCED_IN_ROUTES_SQL)
    if update_secondary_data:
        update_secondary_data_copies(db_conn)


def remove_dangling_shapes(db_conn):
    """
    Remove dangling entries from the shapes directory.

    Parameters
    ----------
    db_conn: sqlite3.Connection
        connection to the GTFS object
    """
    db_conn.execute(DELETE_SHAPES_NOT_REFERENCED_IN_TRIPS_SQL)
    SELECT_MIN_MAX_SHAPE_BREAKS_BY_TRIP_I_SQL = \
        "SELECT trips.trip_I, shape_id, min(shape_break) as min_shape_break, max(shape_break) as max_shape_break FROM trips, stop_times WHERE trips.trip_I=stop_times.trip_I GROUP BY trips.trip_I"
    trip_min_max_shape_seqs= pandas.read_sql(SELECT_MIN_MAX_SHAPE_BREAKS_BY_TRIP_I_SQL, db_conn)

    rows = []
    for row in trip_min_max_shape_seqs.itertuples():
        shape_id, min_shape_break, max_shape_break = row.shape_id, row.min_shape_break, row.max_shape_break
        if min_shape_break is None or max_shape_break is None:
            min_shape_break = float('-inf')
            max_shape_break = float('-inf')
        rows.append( (shape_id, min_shape_break, max_shape_break) )
    DELETE_SQL_BASE = "DELETE FROM shapes WHERE shape_id=? AND (seq<? OR seq>?)"
    db_conn.executemany(DELETE_SQL_BASE, rows)
    remove_dangling_shapes_references(db_conn)


def remove_dangling_shapes_references(db_conn):
    remove_danging_shapes_references_sql = \
        "UPDATE trips SET shape_id=NULL WHERE trips.shape_id NOT IN (SELECT DISTINCT shape_id FROM shapes)"
    db_conn.execute(remove_danging_shapes_references_sql)


def _split_trip(copy_db_conn, orig_trip_I, stop_times_within_buffer_df):
    blocks = []
    next_block = []
    for row in stop_times_within_buffer_df.itertuples():
        if row.within:
            next_block.append(row.seq)
        else:
            if len(next_block) > 1:
                blocks.append(next_block)
            next_block = []
    if len(next_block) > 1:
        blocks.append(next_block)
    orig_trip_df = pandas.read_sql("SELECT * FROM trips WHERE trip_I={trip_I}".format(trip_I=orig_trip_I), copy_db_conn)
    orig_trip_dict = orig_trip_df.to_dict(orient="records")[0]
    for i, seq_block in enumerate(blocks):
        # create new trip for each block,
        # with mostly same trip information as the original
        trip_id_generated = orig_trip_dict['trip_id'] + "_splitted_part_" + str(i)
        insert_generated_trip_sql = \
            "INSERT INTO trips (trip_id, route_I, service_I, direction_id, " \
            "shape_id, headsign, start_time_ds, end_time_ds) " \
            "VALUES (?, ?, ?, ?, ?, ?, ?, ?)"
        values = [trip_id_generated, orig_trip_dict['route_I'],
                  orig_trip_dict['service_I'], orig_trip_dict['direction_id'],
                  None, orig_trip_dict['headsign'], None, None]
        copy_db_conn.execute(insert_generated_trip_sql, values)
        block_trip_I = copy_db_conn.execute("SELECT trip_I from trips WHERE trips.trip_id=?",
                                            [trip_id_generated]).fetchone()[0]
        # alter the trip_I values in the stop_times table for
        seq_values_to_update_str = "(" + ",".join(str(seq) for seq in seq_block) + ")"
        stop_times_update_sql = \
            "UPDATE stop_times SET trip_I={trip_I_generated} WHERE trip_I={orig_trip_I} AND seq IN {seq_block}".format(
                trip_I_generated=block_trip_I,
                orig_trip_I=orig_trip_I,
                seq_block=seq_values_to_update_str
            )
        copy_db_conn.execute(stop_times_update_sql)

    copy_db_conn.execute("DELETE FROM trips WHERE trip_I={orig_trip_I}".format(orig_trip_I=orig_trip_I))
    copy_db_conn.execute("DELETE from stop_times WHERE trip_I={orig_trip_I}".format(orig_trip_I=orig_trip_I))
    copy_db_conn.execute("DELETE FROM shapes WHERE shape_id IN "
                         " (SELECT DISTINCT shapes.shape_id FROM shapes, trips "
                         "     WHERE trip_I={orig_trip_I} AND shapes.shape_id=trips.shape_id)"
                         .format(orig_trip_I=orig_trip_I))

def update_secondary_data_copies(db_conn):
    G = gtfspy.gtfs.GTFS(db_conn)
    G.set_current_process_time_zone()
    update_trip_travel_times_ds(db_conn)
    resequence_stop_times_seq_values(db_conn)
    recreate_days_table(db_conn)
    recreate_day_trips2_table(db_conn)



import networkx
import pandas as pd
from math import isnan
from gtfspy import route_types
from gtfspy.util import wgs84_distance
from warnings import warn

ALL_STOP_TO_STOP_LINK_ATTRIBUTES = [
    "capacity_estimate", "duration_min", "duration_max",
    "duration_median", "duration_avg", "n_vehicles", "route_types",
    "d", "distance_shape",
    "route_I_counts"
]

DEFAULT_STOP_TO_STOP_LINK_ATTRIBUTES = [
    "n_vehicles", "duration_avg",
    "d", "route_I_counts"
]

def walk_transfer_stop_to_stop_network(gtfs, max_link_distance=None):
    """
    Construct the walk network.
    If OpenStreetMap-based walking distances have been computed, then those are used as the distance.
    Otherwise, the great circle distances ("d") is used.

    Parameters
    ----------
    gtfs: gtfspy.GTFS
    max_link_distance: int, optional
        If given, all walking transfers with great circle distance longer
        than this limit (expressed in meters) will be omitted.

    Returns
    -------
    net: networkx.DiGraph
        edges have attributes
            d:
                straight-line distance between stops
            d_walk:
                distance along the road/tracks/..
    """
    if max_link_distance is None:
        max_link_distance = 1000
    net = networkx.Graph()
    _add_stops_to_net(net, gtfs.get_table("stops"))
    stop_distances = gtfs.get_table("stop_distances")
    if stop_distances["d_walk"][0] is None:
        osm_distances_available = False
        warn("Warning: OpenStreetMap-based walking distances have not been computed, using euclidean distances instead."
             "Ignore this warning if running unit tests.")
    else:
        osm_distances_available = True

    for stop_distance_tuple in stop_distances.itertuples():
        from_node = stop_distance_tuple.from_stop_I
        to_node = stop_distance_tuple.to_stop_I

        if osm_distances_available:
            if stop_distance_tuple.d_walk > max_link_distance or isnan(stop_distance_tuple.d_walk):
                continue
            data = {'d': stop_distance_tuple.d, 'd_walk': stop_distance_tuple.d_walk}
        else:
            if stop_distance_tuple.d > max_link_distance:
                continue
            data = {'d': stop_distance_tuple.d}
        net.add_edge(from_node, to_node, data)
    return net


def stop_to_stop_network_for_route_type(gtfs,
                                        route_type,
                                        link_attributes=None,
                                        start_time_ut=None,
                                        end_time_ut=None):
    """
    Get a stop-to-stop network describing a single mode of travel.

    Parameters
    ----------
    gtfs : gtfspy.GTFS
    route_type : int
        See gtfspy.route_types.TRANSIT_ROUTE_TYPES for the list of possible types.
    link_attributes: list[str], optional
        defaulting to use the following link attributes:
            "n_vehicles" : Number of vehicles passed
            "duration_min" : minimum travel time between stops
            "duration_max" : maximum travel time between stops
            "duration_median" : median travel time between stops
            "duration_avg" : average travel time between stops
            "d" : distance along straight line (wgs84_distance)
            "distance_shape" : minimum distance along shape
            "capacity_estimate" : approximate capacity passed through the stop
            "route_I_counts" : dict from route_I to counts
    start_time_ut: int
        start time of the time span (in unix time)
    end_time_ut: int
        end time of the time span (in unix time)

    Returns
    -------
    net: networkx.DiGraph
        A directed graph Directed graph
    """
    if link_attributes is None:
        link_attributes = DEFAULT_STOP_TO_STOP_LINK_ATTRIBUTES
    assert(route_type in route_types.TRANSIT_ROUTE_TYPES)

    stops_dataframe = gtfs.get_stops_for_route_type(route_type)
    net = networkx.DiGraph()
    _add_stops_to_net(net, stops_dataframe)

    events_df = gtfs.get_transit_events(start_time_ut=start_time_ut,
                                        end_time_ut=end_time_ut,
                                        route_type=route_type)
    if len(net.nodes()) < 2:
        assert events_df.shape[0] == 0

    # group events by links, and loop over them (i.e. each link):
    link_event_groups = events_df.groupby(['from_stop_I', 'to_stop_I'], sort=False)
    for key, link_events in link_event_groups:
        from_stop_I, to_stop_I = key
        assert isinstance(link_events, pd.DataFrame)
        # 'dep_time_ut' 'arr_time_ut' 'shape_id' 'route_type' 'trip_I' 'duration' 'from_seq' 'to_seq'
        if link_attributes is None:
            net.add_edge(from_stop_I, to_stop_I)
        else:
            link_data = {}
            if "duration_min" in link_attributes:
                link_data['duration_min'] = float(link_events['duration'].min())
            if "duration_max" in link_attributes:
                link_data['duration_max'] = float(link_events['duration'].max())
            if "duration_median" in link_attributes:
                link_data['duration_median'] = float(link_events['duration'].median())
            if "duration_avg" in link_attributes:
                link_data['duration_avg'] = float(link_events['duration'].mean())
            # statistics on numbers of vehicles:
            if "n_vehicles" in link_attributes:
                link_data['n_vehicles'] = int(link_events.shape[0])
            if "capacity_estimate" in link_attributes:
                link_data['capacity_estimate'] = route_types.ROUTE_TYPE_TO_APPROXIMATE_CAPACITY[route_type] \
                                                 * int(link_events.shape[0])
            if "d" in link_attributes:
                from_lat = net.node[from_stop_I]['lat']
                from_lon = net.node[from_stop_I]['lon']
                to_lat = net.node[to_stop_I]['lat']
                to_lon = net.node[to_stop_I]['lon']
                distance = wgs84_distance(from_lat, from_lon, to_lat, to_lon)
                link_data['d'] = int(distance)
            if "distance_shape" in link_attributes:
                assert "shape_id" in link_events.columns.values
                found = None
                for i, shape_id in enumerate(link_events["shape_id"].values):
                    if shape_id is not None:
                        found = i
                        break
                if found is None:
                    link_data["distance_shape"] = None
                else:
                    link_event = link_events.iloc[found]
                    distance = gtfs.get_shape_distance_between_stops(
                        link_event["trip_I"],
                        int(link_event["from_seq"]),
                        int(link_event["to_seq"])
                    )
                    link_data['distance_shape'] = distance
            if "route_I_counts" in link_attributes:
                link_data["route_I_counts"] = link_events.groupby("route_I").size().to_dict()
            net.add_edge(from_stop_I, to_stop_I, attr_dict=link_data)
    return net


def stop_to_stop_networks_by_type(gtfs):
    """
    Compute stop-to-stop networks for all travel modes (route_types).

    Parameters
    ----------
    gtfs: gtfspy.GTFS

    Returns
    -------
    dict: dict[int, networkx.DiGraph]
        keys should be one of route_types.ALL_ROUTE_TYPES (i.e. GTFS route_types)
    """
    route_type_to_network = dict()
    for route_type in route_types.ALL_ROUTE_TYPES:
        if route_type == route_types.WALK:
            net = walk_transfer_stop_to_stop_network(gtfs)
        else:
            net = stop_to_stop_network_for_route_type(gtfs, route_type)
        route_type_to_network[route_type] = net
    assert len(route_type_to_network) == len(route_types.ALL_ROUTE_TYPES)
    return route_type_to_network

def combined_stop_to_stop_transit_network(gtfs, start_time_ut=None, end_time_ut=None):
    """
    Compute stop-to-stop networks for all travel modes and combine them into a single network.
    The modes of transport are encoded to a single network.
    The network consists of multiple links corresponding to each travel mode.
    Walk mode is not included.

    Parameters
    ----------
    gtfs: gtfspy.GTFS

    Returns
    -------
    net: networkx.MultiDiGraph
        keys should be one of route_types.TRANSIT_ROUTE_TYPES (i.e. GTFS route_types)
    """
    multi_di_graph = networkx.MultiDiGraph()
    for route_type in route_types.TRANSIT_ROUTE_TYPES:
        graph = stop_to_stop_network_for_route_type(gtfs, route_type,
                                                    start_time_ut=start_time_ut, end_time_ut=end_time_ut)
        for from_node, to_node, data in graph.edges(data=True):
            data['route_type'] = route_type
        multi_di_graph.add_edges_from(graph.edges(data=True))
        multi_di_graph.add_nodes_from(graph.nodes(data=True))
    return multi_di_graph

def _add_stops_to_net(net, stops):
    """
    Add nodes to the network from the pandas dataframe describing (a part of the) stops table in the GTFS database.

    Parameters
    ----------
    net: networkx.Graph
    stops: pandas.DataFrame
    """
    for stop in stops.itertuples():
        data = {
            "lat": stop.lat,
            "lon": stop.lon,
            "name": stop.name
        }
        net.add_node(stop.stop_I, data)


def temporal_network(gtfs,
                     start_time_ut=None,
                     end_time_ut=None,
                     route_type=None):
    """
    Compute the temporal network of the data, and return it as a pandas.DataFrame

    Parameters
    ----------
    gtfs : gtfspy.GTFS
    start_time_ut: int | None
        start time of the time span (in unix time)
    end_time_ut: int | None
        end time of the time span (in unix time)
    route_type: int | None
        Specifies which mode of public transport are included, or whether all modes should be included.
        The int should be one of the standard GTFS route_types:
        (see also gtfspy.route_types.TRANSIT_ROUTE_TYPES )
        If route_type is not specified, all modes are included.

    Returns
    -------
    events_df: pandas.DataFrame
        Columns: departure_stop, arrival_stop, departure_time_ut, arrival_time_ut, route_type, route_I, trip_I
    """
    events_df = gtfs.get_transit_events(start_time_ut=start_time_ut,
                                        end_time_ut=end_time_ut,
                                        route_type=route_type)
    events_df.drop('to_seq', 1, inplace=True)
    events_df.drop('shape_id', 1, inplace=True)
    events_df.drop('duration', 1, inplace=True)
    events_df.drop('route_id', 1, inplace=True)
    events_df.rename(
        columns={
            'from_seq': "seq"
        },
        inplace=True
    )
    return events_df


def route_to_route_network(gtfs, walking_threshold, start_time, end_time):
    """
    Creates networkx graph where the nodes are bus routes and a edge indicates that there is a possibility to transfer
    between the routes
    :param gtfs:
    :param walking_threshold:
    :param start_time:
    :param end_time:
    :return:
    """
    graph = networkx.Graph()
    routes = gtfs.get_table("routes")

    for i in routes.itertuples():
        graph.add_node(i.route_id, attr_dict={"type": i.type, "color": route_types.ROUTE_TYPE_TO_COLOR[i.type]})


    query = """SELECT stop1.route_id AS route_id1, stop1.type, stop2.route_id AS route_id2, stop2.type FROM
                (SELECT * FROM stop_distances WHERE d_walk < %s) sd,
                (SELECT * FROM stop_times, trips, routes 
                WHERE stop_times.trip_I=trips.trip_I AND trips.route_I=routes.route_I 
                AND stop_times.dep_time_ds > %s AND stop_times.dep_time_ds < %s) stop1,
                (SELECT * FROM stop_times, trips, routes 
                WHERE stop_times.trip_I=trips.trip_I AND trips.route_I=routes.route_I 
                AND stop_times.dep_time_ds > %s AND stop_times.dep_time_ds < %s) stop2
                WHERE sd.from_stop_I = stop1.stop_I AND sd.to_stop_I = stop2.stop_I AND stop1.route_id != stop2.route_id
                GROUP BY stop1.route_id, stop2.route_id""" % (walking_threshold, start_time, end_time, start_time,
                                                              end_time)
    df = gtfs.execute_custom_query_pandas(query)

    for items in df.itertuples():
        graph.add_edge(items.route_id1, items.route_id2)
    graph.remove_nodes_from(networkx.isolates(graph))
    return graph




# def cluster_network_stops(stop_to_stop_net, distance):
#     """
#     Aggregate graph by grouping nodes that are within a specified distance.
#     The ids of the nodes are tuples of the original stop_Is.
#
#     Parameters
#     ----------
#     network: networkx.DiGraph
#     distance: float
#         group all nodes within this distance.
#
#     Returns
#     -------
#     graph: networkx.Graph
#     """
#     pass


# def aggregate__network(self, graph, distance):
#     """
#     See to_aggregate_line_graph for documentation
#     """
#     raise NotImplementedError("this is not working fully yet")
#     assert distance <= 1000, "only works with distances below 1000 meters"
#     nodes = set(graph.nodes())
#
#     node_distance_graph = networkx.Graph()
#
#     stop_distances = self.get_table("stop_distances")
#     stop_pairs = stop_distances[stop_distances['d'] <= distance]
#     stop_pairs = zip(stop_pairs['from_stop_I'], stop_pairs['to_stop_I'])
#     for node in nodes:
#         node_distance_graph.add_node(node)
#     for node, another_node in stop_pairs:
#         if (node in nodes) and (another_node in nodes):
#             node_distance_graph.add_edge(node, another_node)
#
#     node_group_iter = networkx.connected_components(node_distance_graph)
#
#     aggregate_graph = networkx.Graph()
#     old_node_to_new_node = {}
#     for node_group in node_group_iter:
#         new_node_id = tuple(node for node in node_group)
#         lats = []
#         lons = []
#         names = []
#         for node in node_group:
#             if node not in graph:
#                 # some stops may not part of the original node line graph
#                 # (e.g. if some lines are not considered, or there are extra stops in stops table)
#                 continue
#             old_node_to_new_node[node] = new_node_id
#             lats.append(graph.node[node]['lat'])
#             lons.append(graph.node[node]['lon'])
#             names.append(graph.node[node]['name'])
#         new_lat = numpy.mean(lats)
#         new_lon = numpy.mean(lons)
#         attr_dict = {
#             "lat": new_lat,
#             "lon": new_lon,
#             "names": names
#         }
#         aggregate_graph.add_node(new_node_id, attr_dict=attr_dict)
#
#     for from_node, to_node, data in graph.edges(data=True):
#         new_from_node = old_node_to_new_node[from_node]
#         new_to_node = old_node_to_new_node[to_node]
#         if aggregate_graph.has_edge(new_from_node, new_to_node):
#             edge_data = aggregate_graph.get_edge_data(new_from_node, new_to_node)
#             edge_data['route_ids'].append(data['route_ids'])
#         else:
#             aggregate_graph.add_edge(new_from_node, new_to_node, route_ids=data['route_ids'])
#     return aggregate_graph



from collections import defaultdict

import networkx
import numpy

from gtfspy.routing.connection import Connection
from gtfspy.routing.abstract_routing_algorithm import AbstractRoutingAlgorithm
from gtfspy.routing.node_profile_multiobjective import NodeProfileMultiObjective
from gtfspy.routing.label import merge_pareto_frontiers, LabelTimeWithBoardingsCount, LabelTime, compute_pareto_front, \
    LabelVehLegCount, LabelTimeBoardingsAndRoute, LabelTimeAndRoute
from gtfspy.util import timeit


class MultiObjectivePseudoCSAProfiler(AbstractRoutingAlgorithm):
    """
    Implementation of the profile connection scan algorithm presented in

    http://i11www.iti.uni-karlsruhe.de/extra/publications/dpsw-isftr-13.pdf
    """

    def __init__(self,
                 transit_events,
                 targets,
                 start_time_ut=None,
                 end_time_ut=None,
                 transfer_margin=0,
                 walk_network=None,
                 walk_speed=1.5,
                 verbose=False,
                 track_vehicle_legs=True,
                 track_time=True,
                 track_route=False):
        """
        Parameters
        ----------
        transit_events: list[Connection]
            events are assumed to be ordered in DECREASING departure_time (!)
        targets: int, list
            index of the target stop
        start_time_ut : int, optional
            start time in unixtime seconds
        end_time_ut: int, optional
            end time in unixtime seconds (no connections will be scanned after this time)
        transfer_margin: int, optional
            required extra margin required for transfers in seconds
        walk_speed: float, optional
            walking speed between stops in meters / second.
        walk_network: networkx.Graph, optional
            each edge should have the walking distance as a data attribute ("distance_shape") expressed in meters
        verbose: boolean, optional
            whether to print out progress
        track_vehicle_legs: boolean, optional
            whether to consider the number of vehicle legs
        track_time: boolean, optional
            whether to consider time in the set of pareto_optimal
        """
        AbstractRoutingAlgorithm.__init__(self)
        assert (len(transit_events) == len(set(transit_events))), "Duplicate transit events spotted!"
        self._transit_connections = transit_events
        if start_time_ut is None:
            start_time_ut = transit_events[-1].departure_time
        if end_time_ut is None:
            end_time_ut = transit_events[0].departure_time
        self._start_time = start_time_ut
        self._end_time = end_time_ut
        self._transfer_margin = transfer_margin
        if walk_network is None:
            walk_network = networkx.Graph()
        self._walk_network = walk_network
        self._walk_speed = walk_speed
        self._verbose = verbose

        # algorithm internals

        # initialize stop_profiles
        self._count_vehicle_legs = track_vehicle_legs
        self._consider_time = track_time

        assert(track_time or track_vehicle_legs)
        if track_vehicle_legs:
            if track_time:
                if track_route:
                    self._label_class = LabelTimeBoardingsAndRoute
                else:
                    self._label_class = LabelTimeWithBoardingsCount
            else:
                self._label_class = LabelVehLegCount
        else:
            if track_route:
                self._label_class = LabelTimeAndRoute
            else:
                self._label_class = LabelTime

        print("using label:", str(self._label_class))
        self._stop_departure_times, self._stop_arrival_times = self.__compute_stop_dep_and_arrival_times()
        self._all_nodes = set.union(set(self._stop_departure_times.keys()),
                                    set(self._stop_arrival_times.keys()),
                                    set(self._walk_network.nodes()))

        self._pseudo_connections = self.__compute_pseudo_connections()
        self._add_pseudo_connection_departures_to_stop_departure_times()
        self._all_connections = self._pseudo_connections + self._transit_connections
        self._all_connections.sort(key=lambda connection: (-connection.departure_time, -connection.seq))
        self._augment_all_connections_with_arrival_stop_next_dep_time()
        if isinstance(targets, list):
            self._targets = targets
        else:
            self._targets = [targets]
        self.reset(self._targets)

    @timeit
    def _add_pseudo_connection_departures_to_stop_departure_times(self):
        self._stop_departure_times_with_pseudo_connections = dict(self._stop_departure_times)
        for node in self._all_nodes:
            if node not in self._stop_departure_times_with_pseudo_connections:
                self._stop_departure_times_with_pseudo_connections[node] = list()
        for key, value in self._stop_departure_times_with_pseudo_connections.items():
            self._stop_departure_times_with_pseudo_connections[key] = list(value)
        for pseudo_connection in self._pseudo_connections:
            assert(isinstance(pseudo_connection, Connection))
            self._stop_departure_times_with_pseudo_connections[pseudo_connection.departure_stop]\
                .append(pseudo_connection.departure_time)
        for stop, dep_times in self._stop_departure_times_with_pseudo_connections.items():
            self._stop_departure_times_with_pseudo_connections[stop] = numpy.array(list(sorted(set(dep_times))))


    @timeit
    def __initialize_node_profiles(self):
        self._stop_profiles = dict()
        for node in self._all_nodes:
            walk_duration_to_target = float('inf')
            closest_target = None
            if node in self._targets:
                walk_duration_to_target = 0
                closest_target = node
            else:
                for target in self._targets:
                    if self._walk_network.has_edge(target, node):
                        edge_data = self._walk_network.get_edge_data(target, node)
                        walk_duration = int(edge_data["d_walk"] / float(self._walk_speed))
                        if walk_duration_to_target > walk_duration:
                            walk_duration_to_target = walk_duration
                            closest_target = target

            self._stop_profiles[node] = NodeProfileMultiObjective(dep_times=self._stop_departure_times_with_pseudo_connections[node],
                                                                  label_class=self._label_class,
                                                                  walk_to_target_duration=walk_duration_to_target,
                                                                  transit_connection_dep_times=self._stop_departure_times[node],
                                                                  closest_target=closest_target,
                                                                  node_id=node)
    @timeit
    def __compute_stop_dep_and_arrival_times(self):
        stop_departure_times = defaultdict(lambda: list())
        stop_arrival_times = defaultdict(lambda: list())
        for connection in self._transit_connections:
            stop_arrival_times[connection.arrival_stop].append(connection.arrival_time)
            stop_departure_times[connection.departure_stop].append(connection.departure_time)
        for stop in stop_departure_times:
            stop_departure_times[stop] = numpy.array(sorted(list(set(stop_departure_times[stop]))))
        for stop in stop_arrival_times:
            stop_arrival_times[stop] = numpy.array(sorted(list(set(stop_arrival_times[stop]))))
        return stop_departure_times, stop_arrival_times


    @timeit
    def __compute_pseudo_connections(self):
        print("Started computing pseudoconnections")
        pseudo_connections = []
        # DiGraph makes things iterate both ways (!)
        for u, v, data in networkx.DiGraph(self._walk_network).edges(data=True):
            walk_duration = int(data["d_walk"] / float(self._walk_speed))  # round to one second accuracy
            total_walk_time_with_transfer = walk_duration + self._transfer_margin
            in_times = self._stop_arrival_times[u]
            out_times = self._stop_departure_times[v]
            j = 0
            n_in_times = len(in_times)
            n_out_times = len(out_times)
            if n_in_times == 0 or n_out_times == 0:
                continue
            i = 0
            while i < n_in_times and j < n_out_times:
                if in_times[i] + total_walk_time_with_transfer > out_times[j]:
                    j += 1  # connection j cannot be reached -> need to check next j -> increase out_time
                else:
                    # if next element still satisfies the wanted condition, go on and increase i!
                    while i + 1 < n_in_times and in_times[i + 1] + total_walk_time_with_transfer < out_times[j]:
                        i += 1
                    dep_time = in_times[i]
                    arr_time = out_times[j]
                    from_stop = u
                    to_stop = v
                    waiting_time = arr_time - dep_time - total_walk_time_with_transfer
                    assert(waiting_time >= 0)
                    pseudo = Connection(from_stop, to_stop, arr_time - walk_duration, arr_time,
                                        Connection.WALK_TRIP_ID,
                                        Connection.WALK_SEQ,
                                        is_walk=True)
                    pseudo_connections.append(pseudo)
                    i += 1
        print("Computed pseudoconnections")
        return pseudo_connections

    @timeit
    def _augment_all_connections_with_arrival_stop_next_dep_time(self):
        for connection in self._all_connections:
            assert(isinstance(connection, Connection))
            to_stop = connection.arrival_stop

            arr_stop_dep_times = self._stop_departure_times_with_pseudo_connections[to_stop]

            arr_stop_next_dep_time = float('inf')
            if len(arr_stop_dep_times) > 0:
                if connection.is_walk:
                    index = numpy.searchsorted(arr_stop_dep_times, connection.arrival_time)
                else:
                    index = numpy.searchsorted(arr_stop_dep_times, connection.arrival_time + self._transfer_margin)
                if 0 <= index < len(arr_stop_dep_times):
                    arr_stop_next_dep_time = arr_stop_dep_times[index]
            if connection.is_walk and not (arr_stop_next_dep_time < float('inf')):
                assert (arr_stop_next_dep_time < float('inf'))
            connection.arrival_stop_next_departure_time = arr_stop_next_dep_time

    def _get_modified_arrival_node_labels(self, connection):
        # get all different "accessible" / arrival times (Pareto-optimal sets)
        arrival_profile = self._stop_profiles[connection.arrival_stop]  # NodeProfileMultiObjective
        assert (isinstance(arrival_profile, NodeProfileMultiObjective))

        arrival_node_labels_orig = arrival_profile.evaluate(connection.arrival_stop_next_departure_time,
                                                            first_leg_can_be_walk=not connection.is_walk,
                                                            connection_arrival_time=connection.arrival_time)

        increment_vehicle_count = (self._count_vehicle_legs and not connection.is_walk)
        # TODO: (?) this copying / modification logic should be moved to the Label / ForwardJourney class ?
        arrival_node_labels_modified = self._copy_and_modify_labels(
            arrival_node_labels_orig,
            connection,
            increment_vehicle_count=increment_vehicle_count,
            first_leg_is_walk=connection.is_walk
        )
        if connection.is_walk:
            connection.is_walk = True
        arrival_node_labels_modified = compute_pareto_front(arrival_node_labels_modified)
        return arrival_node_labels_modified

    def _get_trip_labels(self, connection):
        # best labels from this current trip
        if not connection.is_walk:
            trip_labels = self._copy_and_modify_labels(self.__trip_labels[connection.trip_id],
                                                       connection,
                                                       increment_vehicle_count=False,
                                                       first_leg_is_walk=False)
        else:
            trip_labels = list()
        return trip_labels

    @timeit
    def _run(self):
        previous_departure_time = float("inf")
        n_connections_tot = len(self._all_connections)
        for i, connection in enumerate(self._all_connections):
            # basic checking + printing progress:
            if self._verbose and i % 1000 == 0:
                print("\r", i, "/", n_connections_tot, " : ", "%.2f" % round(float(i) / n_connections_tot, 3), end='', flush=True)
            assert (isinstance(connection, Connection))
            assert (connection.departure_time <= previous_departure_time)
            previous_departure_time = connection.departure_time

            # Get labels from the stop (possibly subject to buffer time)
            arrival_node_labels = self._get_modified_arrival_node_labels(connection)
            # This is for the labels staying "in the vehicle"
            trip_labels = self._get_trip_labels(connection)

            # Then, compute Pareto-frontier of these alternatives:
            all_pareto_optimal_labels = merge_pareto_frontiers(arrival_node_labels, trip_labels)

            # Update labels for this trip
            if not connection.is_walk:
                self.__trip_labels[connection.trip_id] = all_pareto_optimal_labels

            # Update labels for the departure stop profile (later: with the sets of pareto-optimal labels)
            self._stop_profiles[connection.departure_stop].update(all_pareto_optimal_labels,
                                                                  connection.departure_time)


        print("finalizing profiles!")
        self._finalize_profiles()

    def _finalize_profiles(self):
        """
        Deal with the first walks by joining profiles to other stops within walking distance.
        """
        for stop, stop_profile in self._stop_profiles.items():
            assert (isinstance(stop_profile, NodeProfileMultiObjective))
            neighbor_label_bags = []
            walk_durations_to_neighbors = []
            departure_arrival_stop_pairs = []
            if stop_profile.get_walk_to_target_duration() != 0 and stop in self._walk_network.node:
                neighbors = networkx.all_neighbors(self._walk_network, stop)
                for neighbor in neighbors:
                    neighbor_profile = self._stop_profiles[neighbor]
                    assert (isinstance(neighbor_profile, NodeProfileMultiObjective))
                    neighbor_real_connection_labels = neighbor_profile.get_labels_for_real_connections()
                    neighbor_label_bags.append(neighbor_real_connection_labels)
                    walk_durations_to_neighbors.append(int(self._walk_network.get_edge_data(stop, neighbor)["d_walk"] /
                                                       self._walk_speed))
                    departure_arrival_stop_pairs.append((stop, neighbor))
            stop_profile.finalize(neighbor_label_bags, walk_durations_to_neighbors, departure_arrival_stop_pairs)

    @property
    def stop_profiles(self):
        """
        Returns
        -------
        _stop_profiles : dict[int, NodeProfileMultiObjective]
            The pareto tuples necessary.
        """
        assert self._has_run
        return self._stop_profiles

    def _copy_and_modify_labels(self, labels, connection, increment_vehicle_count=False, first_leg_is_walk=False):
        if self._label_class == LabelTimeBoardingsAndRoute or self._label_class == LabelTimeAndRoute:
            labels_copy = [label.get_label_with_connection_added(connection) for label in labels]
        else:
            labels_copy = [label.get_copy() for label in labels]

        for label in labels_copy:
            label.departure_time = connection.departure_time
            if self._label_class == LabelTimeAndRoute or self._label_class == LabelTimeBoardingsAndRoute:
                label.movement_duration += connection.duration()
            if increment_vehicle_count:
                label.n_boardings += 1
            label.first_leg_is_walk = first_leg_is_walk

        return labels_copy

    def reset(self, targets):
        if isinstance(targets, list):
            self._targets = targets
        else:
            self._targets = [targets]
        for target in targets:
            assert(target in self._all_nodes)
        self.__initialize_node_profiles()
        self.__trip_labels = defaultdict(lambda: list())
        self._has_run = False



from unittest import TestCase

import networkx
from six import StringIO

from gtfspy.routing.connection import Connection
from gtfspy.routing.label import min_arrival_time_target, LabelTimeWithBoardingsCount, LabelTime
from gtfspy.routing.multi_objective_pseudo_connection_scan_profiler import MultiObjectivePseudoCSAProfiler
from gtfspy.routing.node_profile_multiobjective import NodeProfileMultiObjective

import pyximport
pyximport.install()

class TestMultiObjectivePseudoCSAProfiler(TestCase):
    # noinspection PyAttributeOutsideInit

    def setUp(self):
        event_list_raw_data = [
            (2, 4, 40, 50, "trip_6", 1),
            (1, 3, 32, 40, "trip_5", 1),
            (3, 4, 32, 35, "trip_4", 1),
            (2, 3, 25, 30, "trip_3", 1),
            (1, 2, 10, 20, "trip_2", 1),
            (0, 1, 0, 10, "trip_1", 1)
        ]
        self.transit_connections = list(map(lambda el: Connection(*el), event_list_raw_data))
        self.walk_network = networkx.Graph()
        self.walk_network.add_edge(1, 2, {"d_walk": 20})
        self.walk_network.add_edge(3, 4, {"d_walk": 15})
        self.walk_speed = 1
        self.target_stop = 4
        self.transfer_margin = 0
        self.start_time = 0
        self.end_time = 50

    def test_pseudo_connections(self):
        event_list_raw_data = [
            (0, 1, 10, 20, "trip_6", 1),
            (2, 3, 42, 50, "trip_5", 1)
        ]
        transit_connections = list(map(lambda el: Connection(*el), event_list_raw_data))
        walk_network = networkx.Graph()
        walk_network.add_edge(1, 2, {"d_walk": 20})
        walk_speed = 1
        target_stop = 3
        transfer_margin = 0
        start_time = 0
        end_time = 50
        csa_profile = MultiObjectivePseudoCSAProfiler(transit_connections, target_stop,
                                                      start_time, end_time, transfer_margin,
                                                      walk_network, walk_speed)
        self.assertEqual(len(csa_profile._all_connections), 3)
        pseudo_connection = csa_profile._all_connections[1]
        self.assertTrue(pseudo_connection.is_walk)
        self.assertEqual(pseudo_connection.departure_time, 42 - 20)
        self.assertEqual(pseudo_connection.arrival_time, 42)
        self.assertEqual(pseudo_connection.departure_stop, 1)
        self.assertEqual(pseudo_connection.arrival_stop, 2)

        node_to_connection_dep_times = {
            0: [10],
            1: [42 - 20],
            2: [42],
            3: [],
        }

        for node, dep_times in node_to_connection_dep_times.items():
            profile = csa_profile._stop_profiles[node]
            for dep_time in dep_times:
                self.assertIn(dep_time, profile.dep_times_to_index, "Node: " + str(node))
            for dep_time in profile.dep_times_to_index:
                self.assertIn(dep_time, dep_times, "Node: " + str(node))

        for connection in csa_profile._all_connections:
            arrival_stop_profile = csa_profile._stop_profiles[connection.arrival_stop]
            departure_stop_profile = csa_profile._stop_profiles[connection.departure_stop]
            self.assertIsInstance(arrival_stop_profile, NodeProfileMultiObjective)
            self.assertIsInstance(departure_stop_profile, NodeProfileMultiObjective)
            self.assertIn(connection.departure_time, departure_stop_profile.dep_times_to_index)
            if connection.arrival_stop_next_departure_time != float('inf'):
                self.assertIn(connection.arrival_stop_next_departure_time, arrival_stop_profile.dep_times_to_index)


    def test_pseudo_connections_with_transfer_margin(self):
        event_list_raw_data = [
            (0, 1, 10, 20, "trip_6", 1),
            (2, 3, 42, 50, "trip_5", 1)
        ]
        transit_connections = list(map(lambda el: Connection(*el), event_list_raw_data))
        walk_network = networkx.Graph()
        walk_network.add_edge(1, 2, {"d_walk": 10})
        walk_speed = 1
        target_stop = 3
        transfer_margin = 5
        start_time = 0
        end_time = 50
        csa_profile = MultiObjectivePseudoCSAProfiler(transit_connections, target_stop,
                                                      start_time, end_time, transfer_margin,
                                                      walk_network, walk_speed)
        transfer_connection = csa_profile._all_connections[1]
        self.assertEqual(transfer_connection.arrival_stop, 2)
        self.assertEqual(transfer_connection.arrival_stop_next_departure_time, 42)
        self.assertEqual(transfer_connection.departure_stop, 1)
        self.assertEqual(transfer_connection.departure_time, 42 - 10)
        self.assertEqual(transfer_connection.is_walk, True)
        self.assertEqual(transfer_connection.arrival_time, 42)

    def test_basics(self):
        csa_profile = MultiObjectivePseudoCSAProfiler(self.transit_connections, self.target_stop,
                                                      self.start_time, self.end_time, self.transfer_margin,
                                                      self.walk_network, self.walk_speed)
        csa_profile.run()

        stop_3_labels = csa_profile.stop_profiles[3].get_final_optimal_labels()
        self.assertEqual(len(stop_3_labels), 1)
        self.assertIn(LabelTimeWithBoardingsCount(32, 35, n_boardings=1, first_leg_is_walk=False), stop_3_labels)

        stop_2_labels = csa_profile.stop_profiles[2].get_final_optimal_labels()
        self.assertEqual(len(stop_2_labels), 3)
        self.assertIn(LabelTimeWithBoardingsCount(40, 50, n_boardings=1, first_leg_is_walk=False), stop_2_labels)
        self.assertIn(LabelTimeWithBoardingsCount(25, 35, n_boardings=2, first_leg_is_walk=False), stop_2_labels)
        self.assertIn(LabelTimeWithBoardingsCount(25, 45, n_boardings=1, first_leg_is_walk=False), stop_2_labels)


        stop_one_profile = csa_profile.stop_profiles[1]
        stop_one_pareto_labels = stop_one_profile.get_final_optimal_labels()

        labels = list()
        # these should exist at least:
        labels.append(LabelTimeWithBoardingsCount(departure_time=10, arrival_time_target=35, n_boardings=3, first_leg_is_walk=False))
        labels.append(LabelTimeWithBoardingsCount(departure_time=20, arrival_time_target=50, n_boardings=1, first_leg_is_walk=False))
        labels.append(LabelTimeWithBoardingsCount(departure_time=32, arrival_time_target=55, n_boardings=1, first_leg_is_walk=False))

    def test_multiple_targets(self):
        event_list_raw_data = [
            (1, 4, 40, 50, "trip", 1),
            (1, 5, 30, 40, "trip", 1),
        ]
        transit_connections = list(map(lambda el: Connection(*el), event_list_raw_data))
        walk_network = networkx.Graph()
        walk_speed = 1
        source_stop = 1
        targets = [4, 5]
        transfer_margin = 0
        start_time = 0
        end_time = 60

        csa_profile = MultiObjectivePseudoCSAProfiler(transit_connections, targets,
                                                      start_time, end_time, transfer_margin,
                                                      walk_network, walk_speed)
        csa_profile.run()
        source_stop_profile = csa_profile.stop_profiles[source_stop]
        final_labels = source_stop_profile.get_final_optimal_labels()
        self.assertEqual(2, len(final_labels))

    def test_simple(self):
        event_list_raw_data = [
            (2, 4, 40, 50, "trip_5", 1),
        ]
        transit_connections = list(map(lambda el: Connection(*el), event_list_raw_data))
        walk_network = networkx.Graph()
        walk_network.add_edge(1, 2, {"d_walk": 20})
        walk_network.add_edge(3, 4, {"d_walk": 15})
        walk_speed = 1
        source_stop = 1
        target_stop = 4
        transfer_margin = 0
        start_time = 0
        end_time = 50

        csa_profile = MultiObjectivePseudoCSAProfiler(transit_connections, target_stop,
                                                      start_time, end_time, transfer_margin,
                                                      walk_network, walk_speed)
        csa_profile.run()
        source_stop_profile = csa_profile.stop_profiles[source_stop]
        self.assertTrue(source_stop_profile._finalized)
        self.assertTrue(source_stop_profile._closed)

        source_stop_labels = source_stop_profile.get_final_optimal_labels()

        labels = list()
        labels.append(LabelTimeWithBoardingsCount(departure_time=20,
                                                  arrival_time_target=50,
                                                  n_boardings=1,
                                                  first_leg_is_walk=True))

        self._assert_label_sets_equal(
            labels,
            source_stop_labels
        )

    def test_last_leg_is_walk(self):
        event_list_raw_data = [
            (0, 1, 0, 10, "trip_1", 1)
        ]
        transit_connections = list(map(lambda el: Connection(*el), event_list_raw_data))
        walk_network = networkx.Graph()
        walk_network.add_edge(1, 2, {"d_walk": 20})

        walk_speed = 1
        source_stop = 0
        target_stop = 2
        transfer_margin = 0
        start_time = 0
        end_time = 50
        labels = list()
        labels.append(LabelTimeWithBoardingsCount(departure_time=0, arrival_time_target=30, n_boardings=1, first_leg_is_walk=False))

        csa_profile = MultiObjectivePseudoCSAProfiler(transit_connections, target_stop,
                                                      start_time, end_time, transfer_margin,
                                                      walk_network, walk_speed)
        csa_profile.run()
        found_tuples = csa_profile.stop_profiles[source_stop].get_final_optimal_labels()
        self._assert_label_sets_equal(found_tuples, labels)

    def test_walk_is_faster_than_by_trip(self):
        event_list_raw_data = [
            (0, 1, 0, 10, "trip_1", 1)
        ]
        transit_connections = list(map(lambda el: Connection(*el), event_list_raw_data))
        walk_speed = 0.5
        source_stop = 0
        target_stop = 1
        transfer_margin = 0
        start_time = 0
        end_time = 50

        walk_network = networkx.Graph()
        walk_network.add_edge(0, 1, {"d_walk": 1})
        csa_profile = MultiObjectivePseudoCSAProfiler(transit_connections, target_stop,
                                                      start_time, end_time, transfer_margin,
                                                      walk_network, walk_speed)
        csa_profile.run()
        source_profile = csa_profile.stop_profiles[source_stop]
        self.assertEqual(min_arrival_time_target(source_profile.evaluate(0, first_leg_can_be_walk=True)), 2)
        found_tuples = source_profile.get_final_optimal_labels()
        self.assertEqual(len(found_tuples), 0)

    def test_no_multiple_walks(self):
        event_list_raw_data = [
            (0, 1, 0, 1, "trip_1", 1),
            (1, 0, 0, 1, "trip_2", 1),
            (0, 1, 2, 3, "trip_3", 1),
            (1, 0, 2, 3, "trip_4", 1),
            (0, 1, 4, 5, "trip_5", 1),
            (1, 0, 4, 5, "trip_6", 1),
            (1, 2, 5, 6, "trip_7", 1),
            (2, 1, 5, 6, "trip_8", 1),
            (1, 2, 2, 3, "trip_7", 2),
            (2, 1, 2, 3, "trip_8", 2)
        ]
        transit_connections = list(map(lambda el: Connection(*el), event_list_raw_data))
        walk_network = networkx.Graph()
        walk_network.add_edge(0, 1, {"d_walk": 1})
        walk_network.add_edge(2, 1, {"d_walk": 1})
        walk_speed = 10
        transfer_margin = 0
        start_time = 0
        end_time = 50

        csa_profile = MultiObjectivePseudoCSAProfiler(transit_connections, 2,
                                                      start_time, end_time, transfer_margin,
                                                      walk_network, walk_speed)
        csa_profile.run()
        source_profile = csa_profile.stop_profiles[0]
        print(source_profile.get_final_optimal_labels())
        for label in source_profile.get_final_optimal_labels():
            self.assertGreater(label.n_boardings, 0)

    def test_target_node_not_in_walk_network(self):
        event_list_raw_data = [
            (0, 1, 0, 10, "trip_1", 1)
        ]
        transit_connections = list(map(lambda el: Connection(*el), event_list_raw_data))
        walk_speed = 2
        source_stop = 0
        target_stop = 1
        transfer_margin = 0
        start_time = 0
        end_time = 50

        walk_network = networkx.Graph()
        csa_profile = MultiObjectivePseudoCSAProfiler(transit_connections, target_stop,
                                                      start_time, end_time, transfer_margin,
                                                      walk_network, walk_speed)
        csa_profile.run()
        source_profile = csa_profile.stop_profiles[source_stop]
        self.assertEqual(min_arrival_time_target(source_profile.evaluate(0, 0)), 10)
        found_tuples = source_profile.get_final_optimal_labels()
        self.assertEqual(len(found_tuples), 1)

    def test_pareto_optimality(self):
        event_list_raw_data = [
            (0, 2, 0, 10, "trip_1", 1),
            (0, 1, 2, 5, "trip_2", 1),
            (1, 2, 5, 8, "trip_3", 1)
        ]
        transit_connections = list(map(lambda el: Connection(*el), event_list_raw_data))
        walk_speed = 2
        source_stop = 0
        target_stop = 2
        transfer_margin = 0
        start_time = 0
        end_time = 20
        walk_network = networkx.Graph()
        csa_profile = MultiObjectivePseudoCSAProfiler(transit_connections, target_stop,
                                                      start_time, end_time, transfer_margin,
                                                      walk_network, walk_speed)
        csa_profile.run()
        source_profile = csa_profile.stop_profiles[source_stop]
        self.assertEqual(min_arrival_time_target(source_profile.evaluate(0, 0)), 8)
        found_labels = source_profile.get_final_optimal_labels()
        labels_should_be = list()
        labels_should_be.append(LabelTimeWithBoardingsCount(0, 10, n_boardings=1, first_leg_is_walk=False))
        labels_should_be.append(LabelTimeWithBoardingsCount(2, 8, n_boardings=2, first_leg_is_walk=False))
        self._assert_label_sets_equal(found_labels, labels_should_be)

    def test_transfer_margin(self):
        walk_speed = 1
        target_stop = 2
        start_time = 0
        end_time = 60
        transit_connections = [
            Connection(0, 1, 40, 50, "trip_1", 1),
            Connection(1, 2, 50, 60, "trip_1", 2),
            Connection(3, 1, 40, 50, "trip_2", 1),
        ]
        # case without any transfer margin
        transfer_margin = 0
        csa_profile = MultiObjectivePseudoCSAProfiler(transit_connections, target_stop,
                                                      start_time, end_time, transfer_margin,
                                                      networkx.Graph(), walk_speed)
        csa_profile.run()
        stop_profile_1 = csa_profile.stop_profiles[1]
        stop_profile_3 = csa_profile.stop_profiles[3]
        self.assertEqual(1, len(stop_profile_1.get_final_optimal_labels()))
        self.assertEqual(1, len(stop_profile_3.get_final_optimal_labels()))

        # case with transfer margin
        transfer_margin = 1
        csa_profile = MultiObjectivePseudoCSAProfiler(transit_connections, target_stop,
                                                      start_time, end_time, transfer_margin,
                                                      networkx.Graph(), walk_speed)
        csa_profile.run()
        stop_profile_3 = csa_profile.stop_profiles[3]
        stop_profile_1 = csa_profile.stop_profiles[1]
        self.assertEqual(0, len(stop_profile_3.get_final_optimal_labels()))
        self.assertEqual(1, len(stop_profile_1.get_final_optimal_labels()))

    def test_possible_transfer_margin_bug_with_multiple_arrivals(self):
        walk_speed = 1
        target_stop = 3
        start_time = 0
        end_time = 200
        transfer_margin = 2
        transit_connections = [
            Connection(0, 1, 100, 101, "trip_0", 1),
            Connection(4, 1, 102, 104, "trip_1", 1),
            Connection(2, 3, 106, 108, "trip_2", 1)
        ]
        walk_network = networkx.Graph()
        walk_network.add_edge(1, 2, {"d_walk": 1})
        csa_profile = MultiObjectivePseudoCSAProfiler(transit_connections, target_stop,
                                                      start_time, end_time, transfer_margin,
                                                      walk_network, walk_speed)
        csa_profile.run()
        profile = csa_profile.stop_profiles[4]
        self.assertEqual(len(profile.get_final_optimal_labels()), 0)
        profile = csa_profile.stop_profiles[0]
        self.assertEqual(len(profile.get_final_optimal_labels()), 1)

    def test_transfer_margin_with_walk(self):
        walk_speed = 1
        target_stop = 3
        start_time = 0
        end_time = 2000
        transit_connections = [
            Connection(0, 1, 1000, 1010, "trip__2", 1),
            Connection(0, 1, 1010, 1020, "trip__1", 1),
            Connection(0, 1, 1020, 1030, "trip_0", 1),
            Connection(0, 1, 1000, 1010, "trip_1", 1),
            Connection(0, 1, 1010, 1020, "trip_2", 1),
            Connection(0, 1, 1020, 1030, "trip_3", 1),
            Connection(0, 1, 1030, 1040, "trip_4", 1),
            Connection(2, 3, 1060, 1070, "trip_6", 1),
        ]

        walk_network = networkx.Graph()
        walk_network.add_edge(1, 2, {"d_walk": 5})
        transfer_margins = [10, 20, 30, 40, 0]
        journey_dep_times = [1030, 1020, 1010, 1000, 1030]

        for transfer_margin, dep_time in zip(transfer_margins, journey_dep_times):
            csa_profile = MultiObjectivePseudoCSAProfiler(transit_connections, target_stop,
                                                          start_time, end_time, transfer_margin,
                                                          walk_network, walk_speed)
            csa_profile.run()
            profile = csa_profile.stop_profiles[0]
            self.assertEqual(len(profile.get_final_optimal_labels()), 1, "transfer_margin=" + str(transfer_margin))
            label = profile.get_final_optimal_labels()[0]
            self.assertEqual(label.departure_time, dep_time, "transfer_margin=" + str(transfer_margin))

    def test_basics_no_transfer_tracking(self):
        csa_profile = MultiObjectivePseudoCSAProfiler(
            self.transit_connections, self.target_stop,
            self.start_time, self.end_time, self.transfer_margin,
            self.walk_network, self.walk_speed, track_vehicle_legs=False
        )
        csa_profile.run()

        stop_3_pareto_tuples = csa_profile.stop_profiles[3].get_final_optimal_labels()
        self.assertEqual(len(stop_3_pareto_tuples), 1)
        self.assertIn(LabelTime(32., 35.), stop_3_pareto_tuples)

        stop_2_pareto_tuples = csa_profile.stop_profiles[2].get_final_optimal_labels()
        self.assertEqual(len(stop_2_pareto_tuples), 2)
        self.assertIn(LabelTime(40., 50.), stop_2_pareto_tuples)
        self.assertIn(LabelTime(25., 35.), stop_2_pareto_tuples)

        source_stop_profile = csa_profile.stop_profiles[1]
        source_stop_pareto_optimal_tuples = source_stop_profile.get_final_optimal_labels()

        pareto_tuples = list()
        pareto_tuples.append(LabelTime(departure_time=10, arrival_time_target=35))
        pareto_tuples.append(LabelTime(departure_time=20, arrival_time_target=50))
        pareto_tuples.append(LabelTime(departure_time=32, arrival_time_target=55))

        self._assert_label_sets_equal(
            pareto_tuples,
            source_stop_pareto_optimal_tuples
        )

    def test_transfers_only(self):
        event_list_raw_data = [
            (7, 2, 20, 30, "trip_6", 1),
            (2, 4, 40, 50, "trip_5", 1),
        ]
        transit_connections = list(map(lambda el: Connection(*el), event_list_raw_data))
        walk_network = networkx.Graph()
        walk_network.add_edge(1, 2, {"d_walk": 20})
        walk_network.add_edge(3, 4, {"d_walk": 15})
        walk_speed = 1
        target_stop = 4
        transfer_margin = 0
        start_time = 0
        end_time = 50

        csa_profile = MultiObjectivePseudoCSAProfiler(transit_connections, target_stop,
                                                      start_time, end_time, transfer_margin,
                                                      walk_network, walk_speed, track_time=False)
        csa_profile.run()

        stop_to_n_boardings = {
            2: 1,
            7: 2,
            3: 0
        }

        for stop, n_veh_legs in stop_to_n_boardings.items():
            labels = csa_profile.stop_profiles[stop].get_final_optimal_labels()
            self.assertEqual(len(labels), 1)
            self.assertEqual(labels[0].n_boardings, n_veh_legs)


    def test_reset(self):
        walk_speed = 1
        target_stop = 2
        start_time = 0
        end_time = 60
        transfer_margin = 0
        transit_connections = [
            Connection(0, 1, 40, 50, "trip_1", 1),
            Connection(1, 2, 55, 60, "trip_1", 1),
            Connection(3, 1, 40, 60, "trip_2", 1)
        ]
        csa_profile = MultiObjectivePseudoCSAProfiler(transit_connections, target_stop,
                                                      start_time, end_time, transfer_margin,
                                                      networkx.Graph(), walk_speed)
        csa_profile.run()
        nodes = [0, 1, 2, 3]
        label_counts = [1, 1, 0, 0]
        for node, count in zip(nodes, label_counts):
            n_labels = len(csa_profile.stop_profiles[node].get_final_optimal_labels())
            self.assertEqual(n_labels, count)

        target_stops = [1]
        csa_profile.reset(target_stops)
        csa_profile.run()
        label_counts = [1, 0, 0, 1]
        for node, count in zip(nodes, label_counts):
            n_labels = len(csa_profile.stop_profiles[node].get_final_optimal_labels())
            self.assertEqual(n_labels, count)

        # TODO: perform a check for the reinitialization of trip_labels
        # THIS IS NOT YET TESTED but should work at the moment
        # RK 9.1.2017

    def test_550_problem(self):
        # There used to be a problem when working with real unixtimes (c-side floating point number problems),
        # this test is one check for that
        event_data = StringIO(
            "from_stop_I,to_stop_I,dep_time_ut,arr_time_ut,route_type,route_id,trip_I,seq\n" +
            "2198,2247,1475530740,1475530860,3,2550,158249,36\n" +
            "2247,2177,1475530860,1475530980,3,2550,158249,37\n")
        import pandas as pd
        events = pd.read_csv(event_data)
        events.sort_values("dep_time_ut", ascending=False, inplace=True)
        connections = [
            Connection(int(e.from_stop_I), int(e.to_stop_I), int(e.dep_time_ut), int(e.arr_time_ut),
                       int(e.trip_I),
                       int(e.seq))
            for e in events.itertuples()
        ]
        csa_profiler = MultiObjectivePseudoCSAProfiler(connections, 2177,
                                                      0, 1475530860*10, 0,
                                                      networkx.Graph(), 0)

        csa_profiler.run()

        profiles = csa_profiler.stop_profiles
        labels_2198 = profiles[2198].get_final_optimal_labels()
        self.assertEqual(len(labels_2198), 1)
        self.assertEqual(labels_2198[0].duration(), 1475530980 - 1475530740)
        labels_2247 = profiles[2247].get_final_optimal_labels()
        self.assertEqual(len(labels_2247), 1)
        self.assertEqual(labels_2247[0].duration(), 1475530980 - 1475530860)

    def test_transfer_on_same_stop_with_multiple_departures(self):
        walk_speed = 1000
        target_stop = 5
        start_time = 0
        end_time = 60
        transfer_margin = 0
        transit_connections = [
            Connection(0, 4, 30, 40, "trip_1", 1),
            Connection(4, 1, 50, 60, "trip_2", 1),
            Connection(4, 2, 50, 60, "trip_3", 1),
            Connection(4, 3, 50, 60, "trip_4", 1),
            Connection(4, target_stop, 70, 100, "trip_5", 1)
        ]
        csa_profiler = MultiObjectivePseudoCSAProfiler(transit_connections, target_stop,
                                                      start_time, end_time, transfer_margin,
                                                      networkx.Graph(), walk_speed)
        csa_profiler.run()
        profiles = csa_profiler.stop_profiles
        assert(profiles[0].get_final_optimal_labels()[0])
        assert(len(profiles[0].get_final_optimal_labels()) > 0)

    def test_transfer_connections_do_not_affect_transfers(self):
        walk_speed = 1000
        target_stop = 1233412
        start_time = 0
        end_time = 60
        transfer_margin = 0
        transit_connections = [
            Connection(0, 1, 30, 40, "trip_1", 1),
            Connection(3, 4, 45, 50, "trip_2", 1),
            Connection(4, 3, 45, 50, "trip_3", 1),
            Connection(5, 3, 45, 50, "trip_4", 1),
            Connection(1, target_stop, 70, 100, "trip_5", 1)
        ]
        walk_network = networkx.Graph()
        walk_network.add_edge(1, 3, {"d_walk": 1})
        walk_network.add_edge(1, 4, {"d_walk": 1})
        walk_network.add_edge(1, 5, {"d_walk": 1})
        csa_profiler = MultiObjectivePseudoCSAProfiler(transit_connections, target_stop,
                                                       start_time, end_time, transfer_margin,
                                                       walk_network, walk_speed)
        csa_profiler.run()
        profiles = csa_profiler.stop_profiles
        assert(profiles[0].get_final_optimal_labels()[0])
        assert(len(profiles[0].get_final_optimal_labels()) > 0)


    def test_transfer_connections_do_not_affect_transfers2(self):
        walk_speed = 1
        target_stop = 0
        start_time = 0
        end_time = 60
        transfer_margin = 0
        transit_connections = [
            Connection(3, 0, 10, 11, "trip_1", 1),
            Connection(2, 1, 5, 6, "trip_2", 1),
            Connection(4, 3, 0, 1, "trip_3", 1)
        ]
        walk_network = networkx.Graph()
        walk_network.add_edge(2, 3, {"d_walk": 1})
        walk_network.add_edge(1, 0, {"d_walk": 1})
        csa_profiler = MultiObjectivePseudoCSAProfiler(transit_connections, target_stop,
                                                       start_time, end_time, transfer_margin,
                                                       walk_network, walk_speed)
        csa_profiler.run()
        profiles = csa_profiler.stop_profiles
        assert(len(profiles[4].get_final_optimal_labels()) == 1)
        optimal_label = profiles[4].get_final_optimal_labels()[0]
        self.assertEqual(optimal_label.departure_time, 0)
        self.assertEqual(optimal_label.arrival_time_target, 7)
        self.assertEqual(optimal_label.n_boardings, 2)

    def test_transfer_connections_do_not_affect_transfers3(self):
        walk_speed = 1
        target_stop = 0
        start_time = 0
        end_time = 60
        transfer_margin = 0
        transit_connections = [
            Connection(3, 0, 10, 11, "t1", 1),
            Connection(2, 1, 5, 6, "t2", 1),
            Connection(7, 2, 3, 4, "tX", 1),
            Connection(5, 6, 2, 3, "--", 1),
            Connection(4, 3, 0, 1, "t3", 1)
        ]

        walk_network = networkx.Graph()
        walk_network.add_edge(7, 3, {"d_walk": 1})
        walk_network.add_edge(1, 0, {"d_walk": 1})
        walk_network.add_edge(5, 3, {"d_walk": 1})
        csa_profiler = MultiObjectivePseudoCSAProfiler(transit_connections, target_stop,
                                                       start_time, end_time, transfer_margin,
                                                       walk_network, walk_speed)
        csa_profiler.run()
        profiles = csa_profiler.stop_profiles
        print(profiles[4].get_final_optimal_labels()[0])
        optimal_labels = profiles[4].get_final_optimal_labels()
        assert(len(optimal_labels) == 2)
        boardings_to_arr_time = {}
        for label in optimal_labels:
            boardings_to_arr_time[label.n_boardings] = label.arrival_time_target
        self.assertEqual(boardings_to_arr_time[2], 11)
        self.assertEqual(boardings_to_arr_time[3], 7)

    def _assert_label_sets_equal(self, found_tuples, should_be_tuples):
        self.assertEqual(len(found_tuples), len(should_be_tuples))
        for found_tuple in found_tuples:
            self.assertIn(found_tuple, should_be_tuples)
        for should_be_tuple in should_be_tuples:
            self.assertIn(should_be_tuple, found_tuples)

    def test_stored_route(self):
        # TODO:
        # - test with multiple targets
        # - test with continuing route
        # - test that timestamps for label and the connection objects match
        csa_profile = MultiObjectivePseudoCSAProfiler(self.transit_connections, self.target_stop,
                                                      self.start_time, self.end_time, self.transfer_margin,
                                                      self.walk_network, self.walk_speed, track_route=True)
        csa_profile.run()
        for stop, profile in csa_profile.stop_profiles.items():
            for bag in profile._label_bags:
                for label in bag:
                    # print(stop, label)
                    cur_label = label
                    journey_legs = []
                    while True:
                        connection = cur_label.connection
                        if isinstance(connection, Connection):
                            journey_legs.append(connection)
                        if not cur_label.previous_label:
                            break
                        cur_label = cur_label.previous_label
                    route_tuples_list = [(x.departure_stop, x.arrival_stop) for x in journey_legs]
                    # print(route_tuples_list)
                    # test that all legs are unique
                    self.assertEqual(len(route_tuples_list), len(set(route_tuples_list)))
                    prev_arr_node = None
                    for route_tuple in route_tuples_list:
                        dep_node = route_tuple[0]
                        arr_node = route_tuple[1]
                        # test that all legs have unique departure and arrival nodes
                        self.assertNotEqual(dep_node, arr_node)
                        if prev_arr_node:
                            # test that legs form an continuous path
                            self.assertEqual(prev_arr_node, dep_node)
                        prev_arr_node = arr_node

    def test_target_self_loops(self):
        event_list_raw_data = [
            (3, 1, 30, 40, "trip_3", 1)
        ]
        transit_connections = list(map(lambda el: Connection(*el), event_list_raw_data))
        walk_network = networkx.Graph()
        walk_network.add_edge(1, 3, {"d_walk": 11})
        walk_speed = 1
        target_stop = 1
        transfer_margin = 0
        start_time = 0
        end_time = 50
        print(walk_network.edges())
        print(transit_connections)
        csa_profile = MultiObjectivePseudoCSAProfiler(transit_connections, target_stop,
                                                      start_time, end_time, transfer_margin,
                                                      walk_network, walk_speed, track_vehicle_legs=True,
                                                      track_time=True, track_route=True)
        csa_profile.run()
        for stop, profile in csa_profile.stop_profiles.items():
            if stop == target_stop:
                self.assertEqual(len(profile.get_final_optimal_labels()), 0)

    def test_journeys_using_movement_duration(self):
        def unpack_route_from_labels(cur_label):
            route = []
            last_arrival_stop = None
            while True:
                connection = cur_label.connection
                if isinstance(connection, Connection):
                    route.append(connection.departure_stop)

                if not cur_label.previous_label:
                    break
                cur_label = cur_label.previous_label
                if isinstance(connection, Connection):
                    last_arrival_stop = connection.arrival_stop
            route.append(last_arrival_stop)
            return route

        event_list_raw_data = [
            (1, 2, 0, 10, "trip_1", 1),
            (2, 3, 10, 20, "trip_1", 1),
            (4, 5, 30, 40, "trip_2", 1),

        ]
        transit_connections = list(map(lambda el: Connection(*el), event_list_raw_data))
        walk_network = networkx.Graph()
        walk_network.add_edge(2, 4, {"d_walk": 10})
        walk_network.add_edge(3, 4, {"d_walk": 10})
        walk_speed = 1
        target_stop = 5
        transfer_margin = 0
        start_time = 0
        end_time = 50

        csa_profile = MultiObjectivePseudoCSAProfiler(transit_connections, target_stop,
                                                      start_time, end_time, transfer_margin,
                                                      walk_network, walk_speed, track_vehicle_legs=False,
                                                      track_time=True, track_route=True)
        csa_profile.run()
        for stop, profile in csa_profile.stop_profiles.items():
            for label_bag in profile._label_bags:
                for label in label_bag:
                    print('origin:', stop, 'n_boardings/movement_duration:', label.movement_duration, 'route:', unpack_route_from_labels(label))
        print('optimal labels:')
        for stop, profile in csa_profile.stop_profiles.items():
            for label in profile.get_final_optimal_labels():

                print('origin:', stop, 'n_boardings/movement_duration:', label.movement_duration, 'route:', unpack_route_from_labels(label))
                #if stop == 1:
                    #assert 3 not in unpack_route_from_labels(label)
                # print('origin:', stop, 'n_boardings:', label.n_boardings, 'route:', unpack_route_from_labels(label))

    def test_journeys_using_movement_duration_last_stop_walk(self):
        def unpack_route_from_labels(cur_label):
            route = []
            last_arrival_stop = None
            print(cur_label)
            while True:
                print(cur_label.previous_label)
                connection = cur_label.connection
                if isinstance(connection, Connection):
                    route.append(connection.departure_stop)

                if not cur_label.previous_label:
                    break
                cur_label = cur_label.previous_label
                if isinstance(connection, Connection):
                    last_arrival_stop = connection.arrival_stop
            route.append(last_arrival_stop)
            return route

        event_list_raw_data = [
            (1, 2, 0, 10, "trip_1", 1),
            (2, 3, 10, 20, "trip_2", 1),
            (4, 5, 30, 40, "trip_3", 1),

        ]
        transit_connections = list(map(lambda el: Connection(*el), event_list_raw_data))
        walk_network = networkx.Graph()
        walk_network.add_edge(2, 4, {"d_walk": 10})
        walk_network.add_edge(3, 4, {"d_walk": 10})
        walk_network.add_edge(5, 6, {"d_walk": 10})
        walk_speed = 1
        target_stop = 5
        transfer_margin = 0
        start_time = 0
        end_time = 50

        csa_profile = MultiObjectivePseudoCSAProfiler(transit_connections, target_stop,
                                                      start_time, end_time, transfer_margin,
                                                      walk_network, walk_speed, track_vehicle_legs=False,
                                                      track_time=True, track_route=True)
        csa_profile.run()
        for stop, profile in csa_profile.stop_profiles.items():
            for label_bag in profile._label_bags:
                for label in label_bag:
                    print('origin:', stop,
                          'n_boardings/movement_duration:', label.movement_duration,
                          'route:', unpack_route_from_labels(label))
        print('optimal labels:')
        for stop, profile in csa_profile.stop_profiles.items():
            for label in profile.get_final_optimal_labels():

                print('origin:', stop,
                      'n_boardings/movement_duration:', label.movement_duration,
                      'route:', unpack_route_from_labels(label))
                #if stop == 1:
                    #assert 3 not in unpack_route_from_labels(label)
                # print('origin:', stop, 'n_boardings:', label.n_boardings, 'route:', unpack_route_from_labels(label))


    def test_zero_length_journeys_potential_bug_1(self):
        event_list_raw_data = [
            (0, 1, 0, 0, "trip_1", 0),
            (1, 2, 0, 0, "trip_1", 1)
        ]
        transit_connections = list(map(lambda el: Connection(*el), event_list_raw_data))
        walk_network = networkx.Graph()
        walk_network.add_edge(10, 1, {"d_walk": 20})
        walk_network.add_edge(1, 11, {"d_walk": 20})
        walk_speed = 1
        target_stop = 11
        transfer_margin = 0
        start_time = 0
        end_time = 50
        csa_profiler = MultiObjectivePseudoCSAProfiler(transit_connections, target_stop,
                                                      start_time, end_time, transfer_margin,
                                                      walk_network, walk_speed,
                                                      track_vehicle_legs=True,
                                                      track_time=True,
                                                      track_route=True)
        csa_profiler.run()
        stop_profile_1 = csa_profiler._stop_profiles[1]
        all_labels_stop_profile_1 = [label for label_bag in stop_profile_1._label_bags for label in label_bag]
        for label in all_labels_stop_profile_1:
            self.assertLess(label.n_boardings, 1, "There should at most a walking label when going from 11 to 1 at any "
                                                  "point in time, now one label has " + str(label.n_boardings) +
                                                  " boardings"
            )

    def test_zero_length_journeys_potential_bug(self):
        s = 0
        a = 1
        b = 2
        t = 3

        event_list_raw_data = [
            (s, a, 0, 0, "trip_1", 1),
            (a, b, 0, 0, "trip_1", 2),
            (b, t, 1, 2, "trip_2", 0)
        ]
        transit_connections = list(map(lambda el: Connection(*el), event_list_raw_data))
        walk_network = networkx.Graph()
        walk_speed = 1
        target_stop = t
        transfer_margin = 0
        start_time = 0
        end_time = 50
        csa_profiler = MultiObjectivePseudoCSAProfiler(transit_connections, target_stop,
                                                       start_time, end_time, transfer_margin,
                                                       walk_network, walk_speed,
                                                       track_vehicle_legs=True,
                                                       track_time=True,
                                                       track_route=True)
        csa_profiler.run()
        stop_profile_a_labels = csa_profiler.stop_profiles[a].get_final_optimal_labels()
        stop_profile_s_labels = csa_profiler.stop_profiles[s].get_final_optimal_labels()
        self.assertEqual(len(stop_profile_a_labels), 1)
        self.assertEqual(len(stop_profile_s_labels), 1)




import os
import unittest
import numpy as np

from gtfspy.gtfs import GTFS
from gtfspy.geometry import get_convex_hull_coordinates, get_approximate_convex_hull_area_km2, approximate_convex_hull_area, compute_buffered_area_of_stops

class GeometryTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """ This method is run once before executing any tests"""
        cls.gtfs_source_dir = os.path.join(os.path.dirname(__file__), "test_data")
        cls.G = GTFS.from_directory_as_inmemory_db(cls.gtfs_source_dir)

    def setUp(self):
        """This method is run once before _each_ test method is executed"""
        self.gtfs = GTFS.from_directory_as_inmemory_db(self.gtfs_source_dir)

    def test_get_convex_hull_coordinates(self):
        lons, lats = get_convex_hull_coordinates(self.gtfs)
        self.assertEqual(len(lons), 5 + 1)
        self.assertEqual(len(lats), 5 + 1)
        for value in lons + lats:
            self.assertIsInstance(value, float)

    def test_get_convex_hull_area(self):
        area = get_approximate_convex_hull_area_km2(self.gtfs)
        print(area)
        self.assertGreater(area, 10)

    def test_approximate_convex_hull_area(self):
        # helsinki railway station, Helsinki
        # leppavaara station, Helsinki
        # pasila railway station, Helsinki
        leppavaara_coords = 60.219163, 24.813390
        pasila_coords = 60.199136, 24.934090
        main_railway_station_coords = 60.171545, 24.940734
        # lat, lon
        lats, lons = list(zip(leppavaara_coords, pasila_coords, main_railway_station_coords))
        approximate_reference = 9.91  # computed using https://asiointi.maanmittauslaitos.fi/karttapaikka/
        computed = approximate_convex_hull_area(lons, lats)
        self.assertTrue(approximate_reference * 0.9  < computed < approximate_reference * 1.1)

    def test_approximate_convex_hull_area(self):
        # helsinki railway station, Helsinki
        # leppavaara station, Helsinki
        # pasila railway station, Helsinki
        leppavaara_coords = 60.219163, 24.813390
        pasila_coords = 60.199136, 24.934090
        main_railway_station_coords = 60.171545, 24.940734
        # lat, lon
        lats, lons = list(zip(leppavaara_coords, pasila_coords, main_railway_station_coords))
        approximate_reference = 9.91  # computed using https://asiointi.maanmittauslaitos.fi/karttapaikka/
        computed = approximate_convex_hull_area(lons, lats)
        self.assertTrue(approximate_reference * 0.9  < computed < approximate_reference * 1.1)

    def test_get_buffered_area_of_stops(self):
        # stop1 is far from stop2, theres no overlap
        # stop1 and stop3 are close and could have overlap
        # The area has an accuracy between 95%-99% of the real value.
        stop1_coords = 61.129094,24.027896
        stop2_coords = 61.747408,23.924279
        stop3_coords = 61.129621,24.027363
        #lat, lon
        lats_1, lons_1 = list(zip(stop1_coords))
        lats_1_2, lons_1_2 = list(zip(stop1_coords, stop2_coords))
        lats_1_3, lons_1_3 = list(zip(stop1_coords, stop3_coords))
        
        #One point buffer
        buffer_onepoint = 100 #100 meters of radius
        true_area = 10000 * np.pi #area = pi * square radius
        area_1 = compute_buffered_area_of_stops(lats_1, lons_1, buffer_onepoint) 
        confidence = true_area * 0.95
        self.assertTrue(confidence < area_1 < true_area)
        
        
        # Two points buffer non-overlap
        # Note: the points are "far away" to avoid overlap, but since they are points in the same city
        # a "really big buffer" could cause overlap and the test is going fail.
        buffer_nonoverlap = 100 #100 meters of radius
        two_points_nonoverlap_true_area = 2 * buffer_nonoverlap ** 2 * np.pi #area = pi * square radius
        area_1_2 = compute_buffered_area_of_stops(lats_1_2, lons_1_2, buffer_nonoverlap) 
        confidence_2 = two_points_nonoverlap_true_area * 0.95
        self.assertTrue(confidence_2 < area_1_2  and area_1_2 < two_points_nonoverlap_true_area)
        
        # Two points buffer with overlap
        # Points so close that will overlap with a radius of 100 meters
        buffer_overlap = 100  # 100 meters of radius
        area_1_3 = compute_buffered_area_of_stops(lats_1_3, lons_1_3, buffer_overlap)
        self.assertLess(area_1, area_1_3)
        self.assertLess(area_1_3, two_points_nonoverlap_true_area)

        # 'Half-overlap'
        from gtfspy.util import wgs84_distance
        lat1, lat3 = lats_1_3
        lon1, lon3 = lons_1_3

        distance = wgs84_distance(lat1, lon1, lat3, lon3)
        # just a little overlap
        buffer = distance / 2. + 1
        area_1_3b = compute_buffered_area_of_stops(lats_1_3, lons_1_3, buffer, resolution=100)
        one_point_true_area =  np.pi * buffer**2
        self.assertLess(one_point_true_area * 1.5, area_1_3b)
        self.assertLess(area_1_3b, 2 * one_point_true_area)

        # no overlap
        buffer = distance / 2. - 1
        area_1_3b = compute_buffered_area_of_stops(lats_1_3, lons_1_3, buffer, resolution=100)
        two_points_nonoverlap_true_area = 2 * buffer ** 2 * np.pi
        self.assertGreater(area_1_3b, two_points_nonoverlap_true_area * 0.95)
        self.assertLess(area_1_3b , two_points_nonoverlap_true_area)



import os
import unittest

from gtfspy.spreading.event import Event
from gtfspy.spreading.spreader import Spreader
from gtfspy.spreading.spreading_stop import SpreadingStop

from gtfspy.gtfs import GTFS


class SpreadingTest(unittest.TestCase):

    @staticmethod
    def test_get_min_visit_time():
        stop_I = 1
        min_transfer_time = 60
        ss = SpreadingStop(stop_I, min_transfer_time)
        assert ss.get_min_visit_time() == float('inf')
        ss.visit_events = [Event(10, 0, stop_I, stop_I, -1)]
        assert ss.get_min_visit_time() == 10
        ss.visit_events.append(Event(5, 0, stop_I, stop_I, -1))
        assert ss.get_min_visit_time() == 5

    @staticmethod
    def test_visit():
        stop_I = 1
        min_transfer_time = 60
        ss = SpreadingStop(stop_I, min_transfer_time)
        e = Event(10, 2, 2, stop_I, 1092)
        assert ss.visit(e)
        assert len(ss.visit_events) == 1
        assert ss.visit_events[0].from_stop_I == 2
        assert ss.visit_events[0].trip_I == 1092
        e2 = Event(66, 1, 2, 1092, stop_I)
        assert ss.visit(e2)
        e3 = Event(5, 1, 2, 1092, stop_I)
        assert ss.visit(e3)
        assert ss.get_min_visit_time() == 5
        e4 = Event(66, 1, 2, 1092, stop_I)
        assert not ss.visit(e4)
        e5 = Event(64, 1, 2, 1092, stop_I)
        assert ss.visit(e5)

    @staticmethod
    def test_can_infect():
        stop_I = 1
        min_transfer_time = 60
        ss = SpreadingStop(stop_I, min_transfer_time)
        trip_I = 2
        e = Event(0, -2, stop_I, 2, trip_I)
        assert not ss.can_infect(e)
        e2 = Event(5, 3, trip_I, 2, trip_I)
        assert not ss.can_infect(e2)
        ss.visit(e2)
        e3 = Event(14, 6, stop_I, stop_I + 1, trip_I)
        assert ss.can_infect(e3)
        e4 = Event(14, 6, stop_I, stop_I + 1, trip_I + 1)
        assert not ss.can_infect(e4)

    @staticmethod
    def test_get_trips():
        gtfs_source_dir = os.path.join(os.path.dirname(__file__), "test_data")
        gtfs = GTFS.from_directory_as_inmemory_db(gtfs_source_dir)
        first_day_start_ut, last_day_start_ut = gtfs.get_day_start_ut_span()
        stop_1_info = gtfs.stop(1)
        lat = stop_1_info.get("lat")
        lon = stop_1_info.get("lon")
        spreader = Spreader(gtfs, first_day_start_ut + 8 * 3600, lat, lon, 24 * 3600, 30, True, 0.5)
        trips = spreader.spread()
        assert isinstance(trips, dict)
        assert "trips" in trips
        assert len(trips["trips"]) > 0
        el = trips["trips"][0]
        keys = "lats lons times route_type name".split()
        for key in keys:
            assert key in el, el





from setuptools import setup, Extension, find_packages

version="0.0.4"

setup(
    name="gtfspy",
    version=version,
    description="Python package for analyzing public transport timetables",
    url="https://github.com/CxAalto/gtfspy",
    packages=find_packages(exclude=["java_routing", "examples"]),
    author="Rainer Kujala",
    author_email="Rainer.Kujala@gmail.com",
    license='MIT',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: GIS',

        # Pick your license as you wish (should match "license" above)
         'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5'
    ], 
    install_requires = [
        "setuptools>=18.0",
        "pandas",
        "networkx==1.11",
        "pyshp",
        "smopy",
        "nose",
        "Cython",
        "six",
        "geoindex",
        "osmread==0.2",
        "shapely",
        "geojson==1.3.5",
        "pyproj",
        "matplotlib-scalebar==0.3.1"
    ],
    ext_modules=[
        Extension(
            'gtfspy.routing.label',
            sources=["gtfspy/routing/label.pyx"],
        ),
    ],
    keywords = ['transit', 'routing' 'gtfs', 'public transport', 'analysis', 'visualization'], # arbitrary keywords
)



from collections import defaultdict

import numpy

from gtfspy.routing.profile_block import ProfileBlock


class ProfileBlockAnalyzer:

    def __init__(self, profile_blocks, cutoff_distance=None, **kwargs):
        """
        Parameters
        ----------
        profile_blocks: list[gtfspy.routing.profile_block.ProfileBlock]
        """
        for i, block in enumerate(profile_blocks[:-1]):
            assert block.start_time < block.end_time
            assert block.end_time == profile_blocks[i + 1].start_time
            assert block.distance_start >= block.distance_end

        self._profile_blocks = profile_blocks
        self._start_time = profile_blocks[0].start_time
        self._end_time = profile_blocks[-1].end_time
        self._cutoff_distance = cutoff_distance
        if cutoff_distance is not None:
            self._apply_cutoff(cutoff_distance)

        self.from_stop_I = None
        self.to_stop_I = None

        for key, value in kwargs.items():
            if key == "from_stop_I":
                self.from_stop_I = value
            if key == "to_stop_I":
                self.to_stop_I = value

    def _apply_cutoff(self, cutoff_distance):
        for block in list(self._profile_blocks):
            block_max = max(block.distance_start, block.distance_end)
            if block_max > cutoff_distance:
                print("applying cutoff")
                blocks = []
                if block.distance_start == block.distance_end or \
                        (block.distance_start > cutoff_distance and block.distance_end > cutoff_distance):
                    blocks.append(
                        ProfileBlock(distance_end=cutoff_distance,
                                     distance_start=cutoff_distance,
                                     start_time=block.start_time,
                                     end_time=block.end_time)
                    )
                else:
                    if (block.distance_end >= cutoff_distance):
                        assert (block.distance_end < cutoff_distance)
                    split_point_x = block.start_time + (block.distance_start - cutoff_distance) / (
                        block.distance_start - block.distance_end) * block.width()
                    if block.distance_start > block.distance_end:
                        start_distance = cutoff_distance
                        end_distance = block.distance_end
                    else:
                        start_distance = block.distance_start
                        end_distance = cutoff_distance
                    first_block = ProfileBlock(block.start_time, split_point_x, start_distance, cutoff_distance)
                    second_block = ProfileBlock(split_point_x, block.end_time, cutoff_distance, end_distance)
                    blocks.append(first_block)
                    blocks.append(second_block)
                index = self._profile_blocks.index(block)
                self._profile_blocks[index:index + 1] = blocks

    def mean(self):
        total_width = self._profile_blocks[-1].end_time - self._profile_blocks[0].start_time
        total_area = sum([block.area() for block in self._profile_blocks])
        return total_area / total_width

    def median(self):
        try:
            distance_split_points_ordered, norm_cdf = self._temporal_distance_cdf()
        except RuntimeError as e:
            return float('inf')

        if len(distance_split_points_ordered) == 0:
            return float('inf')

        left = numpy.searchsorted(norm_cdf, 0.5, side="left")
        right = numpy.searchsorted(norm_cdf, 0.5, side="right")
        if left == len(norm_cdf):
            return float('inf')
        elif left == right:
            left_cdf_val = norm_cdf[right - 1]
            right_cdf_val = norm_cdf[right]
            delta_y = right_cdf_val - left_cdf_val
            assert (delta_y > 0)
            delta_x = (distance_split_points_ordered[right] - distance_split_points_ordered[right - 1])
            median = (0.5 - left_cdf_val) / delta_y * delta_x + distance_split_points_ordered[right - 1]
            return median
        else:
            return distance_split_points_ordered[left]

    def min(self):
        return min([min(block.distance_end, block.distance_start) for block in self._profile_blocks])

    def max(self):
        return max([max(block.distance_end, block.distance_start) for block in self._profile_blocks])

    def largest_finite_distance(self):
        """
        Compute the maximum temporal distance.

        Returns
        -------
        max_temporal_distance : float
        """
        block_start_distances = [block.distance_start for block in self._profile_blocks if
                                 block.distance_start < float('inf')]
        block_end_distances = [block.distance_end for block in self._profile_blocks if
                               block.distance_end < float('inf')]
        distances = block_start_distances + block_end_distances
        if len(distances) > 0:
            return max(distances)
        else:
            return None

    def summary_as_dict(self):
        summary = {"max": self.max(),
                   "min": self.min(),
                   "mean": self.mean(),
                   "median": self.median()}
        if hasattr(self, "from_stop_I"):
            summary['from_stop_I'] = self.from_stop_I
        if hasattr(self, "to_stop_I"):
            summary['to_stop_I'] = self.to_stop_I
        return summary

    def _temporal_distance_cdf(self):
        """
        Temporal distance cumulative density function.

        Returns
        -------
        x_values: numpy.array
            values for the x-axis
        cdf: numpy.array
            cdf values
        """
        distance_split_points = set()
        for block in self._profile_blocks:
            if block.distance_start != float('inf'):
                distance_split_points.add(block.distance_end)
                distance_split_points.add(block.distance_start)

        distance_split_points_ordered = numpy.array(sorted(list(distance_split_points)))
        temporal_distance_split_widths = distance_split_points_ordered[1:] - distance_split_points_ordered[:-1]
        trip_counts = numpy.zeros(len(temporal_distance_split_widths))
        delta_peaks = defaultdict(lambda: 0)

        for block in self._profile_blocks:
            if block.distance_start == block.distance_end:
                delta_peaks[block.distance_end] += block.width()
            else:
                start_index = numpy.searchsorted(distance_split_points_ordered, block.distance_end)
                end_index = numpy.searchsorted(distance_split_points_ordered, block.distance_start)
                trip_counts[start_index:end_index] += 1

        unnormalized_cdf = numpy.array([0] + list(numpy.cumsum(temporal_distance_split_widths * trip_counts)))
        if not (numpy.isclose(
                [unnormalized_cdf[-1]],
                [self._end_time - self._start_time - sum(delta_peaks.values())], atol=1E-4
        ).all()):
            print(unnormalized_cdf[-1], self._end_time - self._start_time - sum(delta_peaks.values()))
            raise RuntimeError("Something went wrong with cdf computation!")

        if len(delta_peaks) > 0:
            for peak in delta_peaks.keys():
                if peak == float('inf'):
                    continue
                index = numpy.nonzero(distance_split_points_ordered == peak)[0][0]
                unnormalized_cdf = numpy.insert(unnormalized_cdf, index, unnormalized_cdf[index])
                distance_split_points_ordered = numpy.insert(distance_split_points_ordered, index,
                                                             distance_split_points_ordered[index])
                # walk_waiting_time_fraction = walk_total_time / (self.end_time_dep - self.start_time_dep)
                unnormalized_cdf[(index + 1):] = unnormalized_cdf[(index + 1):] + delta_peaks[peak]

        norm_cdf = unnormalized_cdf / (unnormalized_cdf[-1] + delta_peaks[float('inf')])
        return distance_split_points_ordered, norm_cdf

    def _temporal_distance_pdf(self):
        """
        Temporal distance probability density function.

        Returns
        -------
        non_delta_peak_split_points: numpy.array
        non_delta_peak_densities: numpy.array
            len(density) == len(temporal_distance_split_points_ordered) -1
        delta_peak_loc_to_probability_mass : dict
        """
        temporal_distance_split_points_ordered, norm_cdf = self._temporal_distance_cdf()
        delta_peak_loc_to_probability_mass = {}

        non_delta_peak_split_points = [temporal_distance_split_points_ordered[0]]
        non_delta_peak_densities = []
        for i in range(0, len(temporal_distance_split_points_ordered) - 1):
            left = temporal_distance_split_points_ordered[i]
            right = temporal_distance_split_points_ordered[i + 1]
            width = right - left
            prob_mass = norm_cdf[i + 1] - norm_cdf[i]
            if width == 0.0:
                delta_peak_loc_to_probability_mass[left] = prob_mass
            else:
                non_delta_peak_split_points.append(right)
                non_delta_peak_densities.append(prob_mass / float(width))
        assert (len(non_delta_peak_densities) == len(non_delta_peak_split_points) - 1)
        return numpy.array(non_delta_peak_split_points), \
               numpy.array(non_delta_peak_densities), delta_peak_loc_to_probability_mass

    def get_vlines_and_slopes_for_plotting(self):
        vertical_lines = []
        slopes = []

        for i, block in enumerate(self._profile_blocks):
            distance_end_minutes = block.distance_end
            distance_start_minutes = block.distance_start

            slope = dict(x=[block.start_time, block.end_time],
                         y=[distance_start_minutes, distance_end_minutes])
            slopes.append(slope)

            if i != 0:
                # no vertical line for the first observation
                previous_duration_minutes = self._profile_blocks[i - 1].distance_end
                vertical_lines.append(dict(x=[block.start_time, block.start_time],
                                           y=[previous_duration_minutes, distance_start_minutes]))
        return vertical_lines, slopes

    def get_blocks(self):
        return self._profile_blocks

    def interpolate(self, time):
        assert(self._start_time <= time <= self._end_time)
        for profile_block in self._profile_blocks:
            # find the first block whose end time is larger than or equal to that of the queried time
            if profile_block.end_time >= time:
                return profile_block.interpolate(time)




import pyximport
pyximport.install()

from unittest import TestCase

from gtfspy.routing.node_profile_multiobjective import NodeProfileMultiObjective
from gtfspy.routing.label import LabelTime, min_arrival_time_target, LabelTimeWithBoardingsCount, LabelVehLegCount


class TestNodeProfileMultiObjective(TestCase):

    def test_evaluate(self):
        node_profile = NodeProfileMultiObjective(dep_times=[3, 1], label_class=LabelTime)

        node_profile.update([LabelTime(departure_time=3, arrival_time_target=4)])
        self.assertEquals(4, min_arrival_time_target(node_profile.evaluate(3)))

        node_profile.update([LabelTime(departure_time=1, arrival_time_target=1)])
        self.assertEquals(1, min_arrival_time_target(node_profile.evaluate(1)))

    def test_pareto_optimality2(self):
        node_profile = NodeProfileMultiObjective(dep_times=[5, 10], label_class=LabelTime)
        pt2 = LabelTime(departure_time=10, arrival_time_target=35, last_leg_is_walk=False)
        node_profile.update([pt2])
        pt1 = LabelTime(departure_time=5, arrival_time_target=35, last_leg_is_walk=False)
        node_profile.update([pt1])
        self.assertEquals(len(node_profile.get_labels_for_real_connections()), 1)

    def test_identity_profile(self):
        identity_profile = NodeProfileMultiObjective(dep_times=[10])
        identity_profile.update([LabelTimeWithBoardingsCount(10, 10, 0, True)])
        self.assertEqual(10, min_arrival_time_target(identity_profile.evaluate(10, first_leg_can_be_walk=True)))

    def test_walk_duration(self):
        node_profile = NodeProfileMultiObjective(dep_times=[10, 5], walk_to_target_duration=27, label_class=LabelTime)
        self.assertEqual(27, node_profile.get_walk_to_target_duration())
        pt2 = LabelTime(departure_time=10, arrival_time_target=35)
        pt1 = LabelTime(departure_time=5, arrival_time_target=35)
        node_profile.update([pt2])
        node_profile.update([pt1])

    def test_pareto_optimality_with_transfers_and_time(self):
        node_profile = NodeProfileMultiObjective(dep_times=[5, 6, 7])
        pt3 = LabelTimeWithBoardingsCount(departure_time=5, arrival_time_target=45, n_boardings=0, first_leg_is_walk=False)
        pt2 = LabelTimeWithBoardingsCount(departure_time=6, arrival_time_target=40, n_boardings=1, first_leg_is_walk=False)
        pt1 = LabelTimeWithBoardingsCount(departure_time=7, arrival_time_target=35, n_boardings=2, first_leg_is_walk=False)
        self.assertTrue(node_profile.update([pt1]))
        self.assertTrue(node_profile.update([pt2]))
        self.assertTrue(node_profile.update([pt3]))
        self.assertEqual(3, len(node_profile.get_labels_for_real_connections()))

    def test_pareto_optimality_with_transfers_only(self):
        LabelClass = LabelVehLegCount
        node_profile = NodeProfileMultiObjective(dep_times=[5, 6, 7], label_class=LabelClass)
        pt3 = LabelClass(departure_time=5, n_vehicle_legs=0, last_leg_is_walk=False)
        pt2 = LabelClass(departure_time=6, n_vehicle_legs=1, last_leg_is_walk=False)
        pt1 = LabelClass(departure_time=7, n_vehicle_legs=2, last_leg_is_walk=False)
        self.assertTrue(node_profile.update([pt1]))
        self.assertTrue(node_profile.update([pt2]))
        self.assertTrue(node_profile.update([pt3]))
        node_profile.finalize()
        self.assertEqual(1, len(node_profile.get_final_optimal_labels()))

    def test_finalize(self):
        node_profile = NodeProfileMultiObjective(label_class=LabelTimeWithBoardingsCount, dep_times=[10])
        own_label = LabelTimeWithBoardingsCount(departure_time=10, arrival_time_target=20, n_boardings=0, first_leg_is_walk=False)
        self.assertTrue(node_profile.update([own_label]))
        neighbor_label = LabelTimeWithBoardingsCount(departure_time=15, arrival_time_target=18, n_boardings=2, first_leg_is_walk=False)
        assert(len(node_profile.get_labels_for_real_connections()) == 1)
        node_profile.finalize([[neighbor_label]], [3])
        assert (len(node_profile.get_final_optimal_labels()) == 2)
        self.assertTrue(any(map(lambda el: el.departure_time == 12, node_profile.get_final_optimal_labels())))

    def test_same_dep_times_fail_in_init(self):
        with self.assertRaises(AssertionError):
            node_profile = NodeProfileMultiObjective(label_class=LabelTimeWithBoardingsCount, dep_times=[10, 10, 20, 20])


    def test_dep_time_skipped_in_update(self):
        label3 = LabelTimeWithBoardingsCount(departure_time=30, arrival_time_target=20, n_boardings=0,
                                             first_leg_is_walk=False)
        label2 = LabelTimeWithBoardingsCount(departure_time=20, arrival_time_target=20, n_boardings=0,
                                             first_leg_is_walk=False)
        label1 = LabelTimeWithBoardingsCount(departure_time=10, arrival_time_target=20, n_boardings=0,
                                             first_leg_is_walk=False)

        # This should work ok
        node_profile = NodeProfileMultiObjective(label_class=LabelTimeWithBoardingsCount, dep_times=[10, 20, 30])
        node_profile.update([label3])
        node_profile.update([label2])
        node_profile.update([label2])
        node_profile.update([label1])

        # This should fail due to dep time 20 missing in between
        with self.assertRaises(AssertionError):
            node_profile = NodeProfileMultiObjective(label_class=LabelTimeWithBoardingsCount, dep_times=[10, 20, 30])
            node_profile.update([label3])
            node_profile.update([label1])

        # This should fail due to dep time 30 not being the first to deal with
        with self.assertRaises(AssertionError):
            node_profile = NodeProfileMultiObjective(label_class=LabelTimeWithBoardingsCount, dep_times=[10, 20, 30])
            node_profile.update([label2])




import contextlib
import datetime
import io
import math
import os
import zipfile
import shutil
import sys
import tempfile
import time
from math import cos

import networkx
import numpy
import pandas as pd
import shapefile as shp

"""
Various unrelated utility functions.
"""

# Below is a race condition, so do it only on import.
# Is there a portable way to do this?
current_umask = os.umask(0)
os.umask(current_umask)

TORADIANS = 3.141592653589793 / 180.
EARTH_RADIUS = 6378137.


def set_process_timezone(TZ):
    """
    Parameters
    ----------
    TZ: string
    """
    try:
        prev_timezone = os.environ['TZ']
    except KeyError:
        prev_timezone = None
    os.environ['TZ'] = TZ
    time.tzset()  # Cause C-library functions to notice the update.
    return prev_timezone


def wgs84_distance(lat1, lon1, lat2, lon2):
    """Distance (in meters) between two points in WGS84 coord system."""
    dLat = math.radians(lat2 - lat1)
    dLon = math.radians(lon2 - lon1)
    a = (math.sin(dLat / 2) * math.sin(dLat / 2) +
         math.cos(math.radians(lat1)) * math.cos(math.radians(lat2)) *
         math.sin(dLon / 2) * math.sin(dLon / 2))
    c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
    d = EARTH_RADIUS * c
    return d


def wgs84_height(meters):
    return meters / (EARTH_RADIUS * TORADIANS)


def wgs84_width(meters, lat):
    R2 = EARTH_RADIUS * cos(lat * TORADIANS)
    return meters / (R2 * TORADIANS)


# cython implementation of this.  It is called _often_.
try:
    from gtfspy.cutil import wgs84_distance
except ImportError:
    pass

possible_tmpdirs = [
    '/tmp',
    ''
]


@contextlib.contextmanager
def create_file(fname=None, fname_tmp=None, tmpdir=None,
                save_tmpfile=False, keepext=False):
    """Context manager for making files with possibility of failure.

    If you are creating a file, it is possible that the code will fail
    and leave a corrupt intermediate file.  This is especially damaging
    if this is used as automatic input to another process.  This context
    manager helps by creating a temporary filename, your code runs and
    creates that temporary file, and then if no exceptions are raised,
    the context manager will move the temporary file to the original
    filename you intended to open.

    Parameters
    ----------
    fname : str
        Target filename, this file will be created if all goes well
    fname_tmp : str
        If given, this is used as the temporary filename.
    tmpdir : str or bool
        If given, put temporary files in this directory.  If `True`,
        then find a good tmpdir that is not on local filesystem.
    save_tmpfile : bool
        If true, the temporary file is not deleteted if an exception
        is raised.
    keepext : bool, default False
            If true, have tmpfile have same extension as final file.

    Returns (as context manager value)
    ----------------------------------
     fname_tmp: str
        Temporary filename to be used.  Same as `fname_tmp`
        if given as an argument.

    Raises
    ------
    Re-raises any except occuring during the context block.
    """
    # Do nothing if requesting sqlite memory DB.
    if fname == ':memory:':
        yield fname
        return
    if fname_tmp is None:
        # no tmpfile name given - compute some basic info
        basename = os.path.basename(fname)
        root, ext = os.path.splitext(basename)
        dir_ = this_dir = os.path.dirname(fname)
        # Remove filename extension, in case this matters for
        # automatic things itself.
        if not keepext:
            root = root + ext
            ext = ''
        if tmpdir:
            # we should use a different temporary directory
            if tmpdir is True:
                # Find a directory ourself, searching some common
                # places.
                for dir__ in possible_tmpdirs:
                    if os.access(dir__, os.F_OK):
                        dir_ = dir__
                        break
        # Make the actual tmpfile, with our chosen tmpdir, directory,
        # extension.  Set it to not delete automatically, since on
        # success we will move it to elsewhere.
        tmpfile = tempfile.NamedTemporaryFile(
            prefix='tmp-' + root + '-', suffix=ext, dir=dir_, delete=False)
        fname_tmp = tmpfile.name
    try:
        yield fname_tmp
    except Exception as e:
        if save_tmpfile:
            print("Temporary file is '%s'" % fname_tmp)
        else:
            os.unlink(fname_tmp)
        raise
    # Move the file back to the original location.
    try:
        os.rename(fname_tmp, fname)
        # We have to manually set permissions.  tempfile does not use
        # umask, for obvious reasons.
        os.chmod(fname, 0o777 & ~current_umask)
    # 'Invalid cross-device link' - you can't rename files across
    # filesystems.  So, we have to fallback to moving it.  But, we
    # want to move it using tmpfiles also, so that the final file
    # appearing is atomic.  We use... tmpfiles.
    except OSError as e:
        # New temporary file in same directory
        tmpfile2 = tempfile.NamedTemporaryFile(
            prefix='tmp-' + root + '-', suffix=ext, dir=this_dir, delete=False)
        # Copy contents over
        shutil.copy(fname_tmp, tmpfile2.name)
        # Rename new tmpfile, unlink old one on other filesystem.
        os.rename(tmpfile2.name, fname)
        os.chmod(fname, 0o666 & ~current_umask)
        os.unlink(fname_tmp)


def execute(cur, *args):
    """Utility function to print sqlite queries before executing.

    Use instead of cur.execute().  First argument is cursor.

    cur.execute(stmt)
    becomes
    util.execute(cur, stmt)
    """
    stmt = args[0]
    if len(args) > 1:
        stmt = stmt.replace('%', '%%').replace('?', '%r')
        print(stmt % (args[1]))
    return cur.execute(*args)


def to_date_string(date):
    if isinstance(date, numpy.int64) or isinstance(date, int):
        date = str(date)
        date = '%s-%s-%s' % (date[:4], date[4:6], date[6:8])
        return date


def ut_to_utc_datetime_str(time_ut):
    dt = datetime.datetime.utcfromtimestamp(time_ut)
    return dt.strftime("%b %d %Y %H:%M:%S")


def str_time_to_day_seconds(time):
    """
    Converts time strings to integer seconds
    :param time: %H:%M:%S string
    :return: integer seconds
    """
    t = str(time).split(':')
    seconds = int(t[0]) * 3600 + int(t[1]) * 60 + int(t[2])
    return seconds


def day_seconds_to_str_time(ds):
    assert ds >= 0
    hours = ds // 3600
    minutes = (ds - hours * 3600) // 60
    seconds = ds % 60
    return "%02d:%02d:%02d" % (hours, minutes, seconds)


def makedirs(path):
    """
    Create directories if they do not exist, otherwise do nothing.

    Return path for convenience
    """
    if not os.path.isdir(path):
        os.makedirs(path)
    return path


def timeit(method):
    """
    A Python decorator for printing out the execution time for a function.

    Adapted from:
    www.andreas-jung.com/contents/a-python-decorator-for-measuring-the-execution-time-of-methods
    """

    def timed(*args, **kw):
        time_start = time.time()
        result = method(*args, **kw)
        time_end = time.time()
        print('timeit: %r %2.2f sec ' % (method.__name__, time_end - time_start))
        return result

    return timed


def corrupted_zip(zip_path):
    import zipfile
    try:
        zip_to_test = zipfile.ZipFile(zip_path)
        # warning = zip_to_test.testzip()
        # if warning is not None:
        #    return str(warning)
        # else:
        return "ok"
    except:
        return "error"


def source_csv_to_pandas(path, table, read_csv_args=None):
    """
    Parameters
    ----------
    path: str
        path to directory or zipfile
    table: str
        name of table
    read_csv_args:
        string arguments passed to the read_csv function

    Returns
    -------
    df: pandas:DataFrame
    """
    if '.txt' not in table:
        table += '.txt'

    if isinstance(path, dict):
        data_obj = path[table]
        f = data_obj.split("\n")
    else:
        if os.path.isdir(path):
            f = open(os.path.join(path, table))

        else:

            z = zipfile.ZipFile(path)
            for path in z.namelist():
                if table in path:
                    table = path
                    break
            try:
                f = zip_open(z, table)
            except KeyError as e:
                return pd.DataFrame()

    if read_csv_args:
        df = pd.read_csv(**read_csv_args)
    else:
        df = pd.read_csv(f)
    return df


def write_shapefile(data, shapefile_path):
    from numpy import int64
    """
    :param data: list of dicts where dictionary contains the keys lons and lats
    :param shapefile_path: path where shapefile is saved
    :return:
    """

    w = shp.Writer(shp.POLYLINE)  # shapeType=3)

    fields = []
    encode_strings = []

    # This makes sure every geom has all the attributes
    w.autoBalance = 1
    # Create all attribute fields except for lats and lons. In addition the field names are saved for the
    # datastoring phase. Encode_strings stores .encode methods as strings for all fields that are strings
    if not fields:
        for key, value in data[0].items():
            if key != u'lats' and key != u'lons':
                fields.append(key)

                if type(value) == float:
                    w.field(key.encode('ascii'), fieldType='N', size=11, decimal=3)
                    print("float", type(value))
                elif type(value) == int or type(value) == int64:
                    print("int", type(value))

                    # encode_strings.append(".encode('ascii')")
                    w.field(key.encode('ascii'), fieldType='N', size=6, decimal=0)
                else:
                    print("other type", type(value))

                    w.field(key.encode('ascii'))

    for dict_item in data:
        line = []
        lineparts = []
        records = []
        records_string = ''
        for lat, lon in zip(dict_item[u'lats'], dict_item[u'lons']):
            line.append([float(lon), float(lat)])
        lineparts.append(line)
        w.line(parts=lineparts)

        # The shapefile records command is built up as strings to allow a differing number of columns
        for field in fields:
            if records_string:
                records_string += ", dict_item['" + field + "']"
            else:
                records_string += "dict_item['" + field + "']"
        method_string = "w.record(" + records_string + ")"

        # w.record(dict_item['name'], dict_item['agency'], dict_item['agency_name'], dict_item['type'], dict_item['lons'])
        print(method_string)
        eval(method_string)
    w.save(shapefile_path)


# Opening files with Universal newlines is done differently in py3
def zip_open(z, filename):
    if sys.version_info[0] == 2:
        return z.open(filename, 'rU')
    else:
        return io.TextIOWrapper(z.open(filename, 'r'))


def draw_net_using_node_coords(net):
    """
    Plot a networkx.Graph by using the lat and lon attributes of nodes.
    Parameters
    ----------
    net : networkx.Graph
    Returns
    -------
    fig : matplotlib.figure
        the figure object where the network is plotted
    """
    import matplotlib.pyplot as plt
    fig = plt.figure()
    node_coords = {}
    for node, data in net.nodes(data=True):
        node_coords[node] = (data['lon'], data['lat'])
    ax = fig.add_subplot(111)
    networkx.draw(net, pos=node_coords, ax=ax, node_size=50)
    return fig


def make_sure_path_exists(path):
    import os
    import errno
    try:
        os.makedirs(path)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise

def difference_of_pandas_dfs(df_self, df_other, col_names=None):
    """
    Returns a dataframe with all of df_other that are not in df_self, when considering the columns specified in col_names
    :param df_self: pandas Dataframe
    :param df_other: pandas Dataframe
    :param col_names: list of column names
    :return:
    """
    df = pd.concat([df_self, df_other])
    df = df.reset_index(drop=True)
    df_gpby = df.groupby(col_names)
    idx = [x[0] for x in list(df_gpby.groups.values()) if len(x) == 1]
    df_sym_diff = df.reindex(idx)
    df_diff = pd.concat([df_other, df_sym_diff])
    df_diff = df_diff.reset_index(drop=True)
    df_gpby = df_diff.groupby(col_names)
    idx = [x[0] for x in list(df_gpby.groups.values()) if len(x) == 2]
    df_diff = df_diff.reindex(idx)
    return df_diff
